#    Copyright 2013 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
from unittest import mock

from oslo_concurrency import processutils
from oslo_serialization import jsonutils
from oslo_utils import imageutils
from oslo_utils.imageutils import format_inspector

from nova.compute import utils as compute_utils
from nova import exception
from nova import test
from nova.virt import images


class QemuTestCase(test.NoDBTestCase):
    def test_qemu_info_with_bad_path(self):
        self.assertRaises(exception.DiskNotFound,
                          images.qemu_img_info,
                          '/path/that/does/not/exist')

    @mock.patch('oslo_concurrency.processutils.execute')
    @mock.patch.object(os.path, 'exists', return_value=True)
    def test_qemu_info_with_errors(self, path_exists, mock_exec):
        err = processutils.ProcessExecutionError(
            exit_code=1, stderr='No such file or directory')
        mock_exec.side_effect = err
        self.assertRaises(exception.DiskNotFound,
                          images.qemu_img_info,
                          '/fake/path')

    @mock.patch.object(os.path, 'exists', return_value=True)
    @mock.patch('nova.privsep.qemu.unprivileged_qemu_img_info',
                return_value={})
    def test_qemu_info_with_no_errors(self, path_exists,
                                      utils_execute):
        image_info = images.qemu_img_info('/fake/path')
        self.assertTrue(image_info)

    @mock.patch('nova.privsep.qemu.unprivileged_qemu_img_info',
                return_value={})
    def test_qemu_info_with_rbd_path(self, utils_execute):
        # Assert that the use of a RBD URI as the path doesn't raise
        # exception.DiskNotFound
        image_info = images.qemu_img_info('rbd:volume/pool')
        self.assertTrue(image_info)

    @mock.patch.object(compute_utils, 'disk_ops_semaphore')
    @mock.patch('nova.privsep.utils.supports_direct_io', return_value=True)
    @mock.patch.object(processutils, 'execute',
                       side_effect=processutils.ProcessExecutionError)
    def test_convert_image_with_errors(self, mocked_execute, mock_direct_io,
                                       mock_disk_op_sema):
        self.assertRaises(exception.ImageUnacceptable,
                          images.convert_image,
                          '/path/that/does/not/exist',
                          '/other/path/that/does/not/exist',
                          'qcow2',
                          'raw')
        mock_disk_op_sema.__enter__.assert_called_once()

    @mock.patch('oslo_concurrency.processutils.execute')
    @mock.patch.object(os.path, 'exists', return_value=True)
    def test_convert_image_with_prlimit_fail(self, path, mocked_execute):
        mocked_execute.side_effect = \
            processutils.ProcessExecutionError(exit_code=-9)
        exc = self.assertRaises(exception.InvalidDiskInfo,
                                images.qemu_img_info,
                                '/fake/path')
        self.assertIn('qemu-img aborted by prlimits', str(exc))

    @mock.patch('oslo_concurrency.processutils.execute')
    @mock.patch.object(os.path, 'exists', return_value=True)
    def test_qemu_img_info_with_disk_not_found(self, exists, mocked_execute):
        """Tests that the initial os.path.exists check passes but the qemu-img
        command fails because the path is gone by the time the command runs.
        """
        path = '/opt/stack/data/nova/instances/some-uuid/disk'
        stderr = (u"qemu-img: Could not open "
                  "'/opt/stack/data/nova/instances/some-uuid/disk': "
                  "Could not open '/opt/stack/data/nova/instances/some-uuid/"
                  "disk': No such file or directory\n")
        mocked_execute.side_effect = (
            processutils.ProcessExecutionError(
                exit_code=1, stderr=stderr))
        self.assertRaises(exception.DiskNotFound, images.qemu_img_info, path)
        exists.assert_called_once_with(path)
        mocked_execute.assert_called_once()

    @mock.patch.object(images, 'IMAGE_API')
    @mock.patch('nova.virt.images.get_image_format')
    @mock.patch.object(images, 'convert_image',
                       side_effect=exception.ImageUnacceptable)
    @mock.patch.object(images, 'qemu_img_info')
    @mock.patch.object(images, 'fetch')
    def test_fetch_to_raw_errors(self, convert_image, qemu_img_info, fetch,
                                 mock_detect, glance):
        inspector = mock_detect.return_value
        inspector.safety_check.return_value = True
        inspector.__str__.return_value = 'qcow2'
        glance.get.return_value = {'disk_format': 'qcow2'}
        qemu_img_info.backing_file = None
        qemu_img_info.file_format = 'qcow2'
        qemu_img_info.virtual_size = 20
        self.assertRaisesRegex(exception.ImageUnacceptable,
                               'Image href123 is unacceptable.*',
                               images.fetch_to_raw,
                               None, 'href123', '/no/path')

    @mock.patch.object(images, 'IMAGE_API')
    @mock.patch('nova.virt.images.get_image_format')
    @mock.patch.object(images, 'convert_image',
                       side_effect=exception.ImageUnacceptable)
    @mock.patch.object(images, 'qemu_img_info')
    @mock.patch.object(images, 'fetch')
    def test_fetch_to_raw_data_file(self, convert_image, qemu_img_info_fn,
                                    fetch, mock_detect, mock_glance):
        mock_glance.get.return_value = {'disk_format': 'qcow2'}
        inspector = mock_detect.return_value
        inspector.safety_check.return_value = True
        inspector.__str__.return_value = 'qcow2'
        # NOTE(danms): the above test needs the following line as well, as it
        # is broken without it.
        qemu_img_info = qemu_img_info_fn.return_value
        qemu_img_info.backing_file = None
        qemu_img_info.file_format = 'qcow2'
        qemu_img_info.virtual_size = 20
        qemu_img_info.format_specific = {'data': {'data-file': 'somefile'}}
        self.assertRaisesRegex(exception.ImageUnacceptable,
                               'Image href123 is unacceptable.*somefile',
                               images.fetch_to_raw,
                               None, 'href123', '/no/path')

    @mock.patch('nova.virt.images.get_image_format')
    @mock.patch.object(images, 'IMAGE_API')
    @mock.patch('os.rename')
    @mock.patch.object(images, 'qemu_img_info')
    @mock.patch.object(images, 'fetch')
    def test_fetch_to_raw_from_raw(self, fetch, qemu_img_info_fn, mock_rename,
                                   mock_glance, mock_detect):
        # Make sure we support a case where we fetch an already-raw image and
        # qemu-img returns None for "format_specific".
        mock_glance.get.return_value = {'disk_format': 'raw'}
        mock_detect.return_value.__str__.return_value = 'raw'
        qemu_img_info = qemu_img_info_fn.return_value
        qemu_img_info.file_format = 'raw'
        qemu_img_info.backing_file = None
        qemu_img_info.format_specific = None
        images.fetch_to_raw(None, 'href123', '/no/path')
        mock_rename.assert_called_once_with('/no/path.part', '/no/path')

    @mock.patch.object(compute_utils, 'disk_ops_semaphore')
    @mock.patch('nova.privsep.utils.supports_direct_io', return_value=True)
    @mock.patch('oslo_concurrency.processutils.execute')
    def test_convert_image_with_direct_io_support(self, mock_execute,
                                                  mock_direct_io,
                                                  mock_disk_op_sema):
        images._convert_image('source', 'dest', 'in_format', 'out_format',
                              run_as_root=False)
        expected = ('qemu-img', 'convert', '-t', 'none', '-O', 'out_format',
                    '-f', 'in_format', 'source', 'dest')
        mock_disk_op_sema.__enter__.assert_called_once()
        self.assertTupleEqual(expected, mock_execute.call_args[0])

    @mock.patch.object(compute_utils, 'disk_ops_semaphore')
    @mock.patch('nova.privsep.utils.supports_direct_io', return_value=False)
    @mock.patch('oslo_concurrency.processutils.execute')
    def test_convert_image_without_direct_io_support(self, mock_execute,
                                                     mock_direct_io,
                                                     mock_disk_op_sema):
        images._convert_image('source', 'dest', 'in_format', 'out_format',
                              run_as_root=False)
        expected = ('qemu-img', 'convert', '-t', 'writeback',
                    '-O', 'out_format', '-f', 'in_format', 'source', 'dest')
        mock_disk_op_sema.__enter__.assert_called_once()
        self.assertTupleEqual(expected, mock_execute.call_args[0])

    def test_convert_image_vmdk_allowed_list_checking(self):
        info = {'format': 'vmdk',
                'format-specific': {
                    'type': 'vmdk',
                    'data': {
                        'create-type': 'monolithicFlat',
                }}}

        # If the format is not in the allowed list, we should get an error
        self.assertRaises(exception.ImageUnacceptable,
                          images.check_vmdk_image, 'foo',
                          imageutils.QemuImgInfo(jsonutils.dumps(info),
                                                 format='json'))

        # With the format in the allowed list, no error
        self.flags(vmdk_allowed_types=['streamOptimized', 'monolithicFlat',
                                       'monolithicSparse'],
                   group='compute')
        images.check_vmdk_image('foo',
                                imageutils.QemuImgInfo(jsonutils.dumps(info),
                                                       format='json'))

        # With an empty list, allow nothing
        self.flags(vmdk_allowed_types=[], group='compute')
        self.assertRaises(exception.ImageUnacceptable,
                          images.check_vmdk_image, 'foo',
                          imageutils.QemuImgInfo(jsonutils.dumps(info),
                                                 format='json'))

    @mock.patch.object(images, 'IMAGE_API')
    @mock.patch('nova.virt.images.get_image_format')
    @mock.patch.object(images, 'fetch')
    @mock.patch('nova.privsep.qemu.unprivileged_qemu_img_info')
    def test_fetch_checks_vmdk_rules(self, mock_info, mock_fetch, mock_detect,
                                     mock_glance):
        mock_glance.get.return_value = {'disk_format': 'vmdk'}
        inspector = mock_detect.return_value
        inspector.safety_check.return_value = True
        inspector.__str__.return_value = 'vmdk'
        info = {'format': 'vmdk',
                'format-specific': {
                    'type': 'vmdk',
                    'data': {
                        'create-type': 'monolithicFlat',
                }}}
        mock_info.return_value = jsonutils.dumps(info)
        with mock.patch('os.path.exists', return_value=True):
            e = self.assertRaises(exception.ImageUnacceptable,
                                  images.fetch_to_raw, None, 'foo', 'anypath')
            self.assertIn('Invalid VMDK create-type specified', str(e))

    @mock.patch('os.rename')
    @mock.patch.object(images, 'IMAGE_API')
    @mock.patch('oslo_utils.imageutils.format_inspector.get_inspector')
    @mock.patch('nova.virt.images.get_image_format')
    @mock.patch.object(images, 'fetch')
    @mock.patch('nova.privsep.qemu.unprivileged_qemu_img_info')
    def test_fetch_iso_is_raw(
        self, mock_info, mock_fetch, mock_detect_file_format, mock_gi,
        mock_glance, mock_rename):
        mock_glance.get.return_value = {'disk_format': 'iso'}
        inspector = mock_gi.return_value.from_file.return_value
        inspector.safety_check.return_value = True
        inspector.__str__.return_value = 'iso'
        mock_detect_file_format.return_value = inspector
        # qemu-img does not have a parser for iso so it is treated as raw
        info = {
            "virtual-size": 356352,
            "filename": "foo.iso",
            "format": "raw",
            "actual-size": 356352,
            "dirty-flag": False
        }
        mock_info.return_value = jsonutils.dumps(info)
        with mock.patch('os.path.exists', return_value=True):
            images.fetch_to_raw(None, 'foo', 'anypath')
        # Make sure we called info with -f raw for an iso, since qemu-img does
        # not support iso
        mock_info.assert_called_once_with('anypath.part', format=None)
        # Make sure that since we considered this to be a raw file, we did the
        # just-rename-don't-convert path
        mock_rename.assert_called_once_with('anypath.part', 'anypath')

    @mock.patch.object(images, 'IMAGE_API')
    @mock.patch('nova.virt.images.get_image_format')
    @mock.patch.object(images, 'qemu_img_info')
    @mock.patch.object(images, 'fetch')
    def test_fetch_to_raw_inspector(self, fetch, qemu_img_info, mock_detect,
                                    mock_glance):
        # Image claims to be qcow2, is qcow2, but fails safety check, so we
        # abort before qemu-img-info
        mock_glance.get.return_value = {'disk_format': 'qcow2'}
        inspector = mock_detect.return_value
        inspector.safety_check.side_effect = (
            format_inspector.SafetyCheckFailed({}))
        inspector.__str__.return_value = 'qcow2'
        self.assertRaises(exception.ImageUnacceptable,
                          images.fetch_to_raw, None, 'href123', '/no.path')
        qemu_img_info.assert_not_called()
        mock_detect.assert_called_once_with('/no.path.part')
        inspector.safety_check.assert_called_once_with()
        mock_glance.get.assert_called_once_with(None, 'href123')

        # Image claims to be qcow2, is qcow2, passes safety check, so we make
        # it all the way to qemu-img-info
        inspector.safety_check.side_effect = None
        qemu_img_info.side_effect = test.TestingException
        self.assertRaises(test.TestingException,
                          images.fetch_to_raw, None, 'href123', '/no.path')

        # Image claims to be qcow2 in glance, but the image is something else,
        # so we abort before qemu-img-info
        qemu_img_info.reset_mock()
        mock_detect.reset_mock()
        inspector.safety_check.reset_mock()
        mock_detect.return_value.__str__.return_value = 'vmdk'
        self.assertRaises(exception.ImageUnacceptable,
                          images.fetch_to_raw, None, 'href123', '/no.path')
        mock_detect.assert_called_once_with('/no.path.part')
        inspector.safety_check.assert_called_once_with()
        qemu_img_info.assert_not_called()

    @mock.patch.object(images, 'IMAGE_API')
    @mock.patch('nova.virt.images.get_image_format')
    @mock.patch.object(images, 'qemu_img_info')
    @mock.patch.object(images, 'fetch')
    def test_fetch_to_raw_inspector_disabled(self, fetch, qemu_img_info,
                                             mock_gi, mock_glance):
        self.flags(disable_deep_image_inspection=True,
                   group='workarounds')
        qemu_img_info.side_effect = test.TestingException
        self.assertRaises(test.TestingException,
                          images.fetch_to_raw, None, 'href123', '/no.path')
        # If deep inspection is disabled, we should never call the inspector
        mock_gi.assert_not_called()
        # ... and we let qemu-img detect the format itself.
        qemu_img_info.assert_called_once_with('/no.path.part')
        mock_glance.get.assert_not_called()

    @mock.patch.object(images, 'IMAGE_API')
    @mock.patch.object(images, 'qemu_img_info')
    @mock.patch('nova.virt.images.get_image_format')
    def test_fetch_inspect_ami(self, detect, imginfo, glance):
        glance.get.return_value = {'disk_format': 'ami'}
        detect.return_value.__str__.return_value = 'raw'
        self.assertRaises(exception.ImageUnacceptable,
                          images.fetch_to_raw, None, 'href123', '/no.path')
        # Make sure 'ami was translated into 'raw' before we call qemu-img
        imginfo.assert_called_once_with('/no.path.part')

    @mock.patch.object(images, 'IMAGE_API')
    @mock.patch.object(images, 'qemu_img_info')
    @mock.patch('nova.virt.images.get_image_format')
    def test_fetch_inspect_aki(self, detect, imginfo, glance):
        glance.get.return_value = {'disk_format': 'aki'}
        detect.return_value.__str__.return_value = 'raw'
        self.assertRaises(exception.ImageUnacceptable,
                          images.fetch_to_raw, None, 'href123', '/no.path')
        # Make sure 'aki was translated into 'raw' before we call qemu-img
        imginfo.assert_called_once_with('/no.path.part')

    @mock.patch.object(images, 'IMAGE_API')
    @mock.patch.object(images, 'qemu_img_info')
    @mock.patch('nova.virt.images.get_image_format')
    def test_fetch_inspect_ari(self, detect, imginfo, glance):
        glance.get.return_value = {'disk_format': 'ari'}
        detect.return_value.__str__.return_value = 'raw'
        self.assertRaises(exception.ImageUnacceptable,
                          images.fetch_to_raw, None, 'href123', '/no.path')
        # Make sure 'aki was translated into 'raw' before we call qemu-img
        imginfo.assert_called_once_with('/no.path.part')

    @mock.patch.object(images, 'IMAGE_API')
    @mock.patch.object(images, 'qemu_img_info')
    def test_fetch_inspect_unknown_format(self, imginfo, glance):
        glance.get.return_value = {'disk_format': 'commodore-64-disk'}
        self.assertRaises(exception.ImageUnacceptable,
                          images.fetch_to_raw, None, 'href123', '/no.path')
        # Unsupported formats do not make it past deep inspection
        imginfo.assert_not_called()

    @mock.patch.object(images, 'IMAGE_API')
    @mock.patch.object(images, 'qemu_img_info')
    @mock.patch('nova.virt.images.get_image_format')
    def test_fetch_inspect_disagrees_qemu(self, mock_detect, imginfo, glance):
        glance.get.return_value = {'disk_format': 'qcow2'}
        mock_detect.return_value.__str__.return_value = 'qcow2'
        # Glance and inspector think it is a qcow2 file, but qemu-img does not
        # agree.
        imginfo.return_value.data_file = None
        imginfo.return_value.file_format = 'vmdk'
        ex = self.assertRaises(exception.ImageUnacceptable,
                               images.fetch_to_raw,
                               None, 'href123', '/no.path')
        self.assertIn('content does not match disk_format', str(ex))
        imginfo.assert_called_once_with('/no.path.part')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Tests For Scheduler image properties weights.
"""
from unittest import mock

from oslo_utils.fixture import uuidsentinel as uuids

from nova import objects
from nova.scheduler import weights
from nova.scheduler.weights import image_props
from nova import test
from nova.tests.unit.scheduler import fakes

PROP_PC = objects.ImageMeta(
    properties=objects.ImageMetaProps(hw_machine_type='pc'))

PROP_WIN_PC = objects.ImageMeta(
    properties=objects.ImageMetaProps(os_distro='windows',
                                      hw_machine_type='pc'))

PROP_LIN = objects.ImageMeta(
    properties=objects.ImageMetaProps(os_distro='linux'))

PROP_LIN_PC = objects.ImageMeta(
    properties=objects.ImageMetaProps(os_distro='linux', hw_machine_type='pc'))


class ImagePropertiesWeigherTestCase(test.NoDBTestCase):

    def setUp(self):
        super().setUp()
        self.weight_handler = weights.HostWeightHandler()
        self.weighers = [image_props.ImagePropertiesWeigher()]

        self.stub_out('nova.objects.ImageMeta.from_instance',
                      self._fake_image_meta_from_instance)

    @staticmethod
    def _fake_image_meta_from_instance(self, inst):
        # inst1 is on host2
        if inst.uuid == uuids.inst1:
            return PROP_WIN_PC
        # inst2 and inst3 are on host3
        elif inst.uuid == uuids.inst2:
            return PROP_LIN
        elif inst.uuid == uuids.inst3:
            return PROP_LIN_PC
        # inst4 is on host4
        elif inst.uuid == uuids.inst4:
            return PROP_LIN
        else:
            return objects.ImageMeta(properties=objects.ImageMetaProps())

    def _get_weighed_host(self, hosts, request_spec=None):
        if request_spec is None:
            request_spec = objects.RequestSpec(image=None)
        return self.weight_handler.get_weighed_objects(self.weighers,
                hosts, request_spec)[0]

    def _get_all_hosts(self):

        host_values = [
            # host1 has no instances
            ('host1', 'node1', {'instances': {}}),
            #  host2 has one instance with os_distro=win and hw_machine_type=pc
            ('host2', 'node2', {'instances': {uuids.inst1: objects.Instance(
                                                        uuid=uuids.inst1)}}),
            # host3 has two instances
            # (os_distro=lin twice, hw_machine_type=pc only once)
            ('host3', 'node3', {'instances': {uuids.inst2: objects.Instance(
                                                        uuid=uuids.inst2),
                                              uuids.inst3: objects.Instance(
                                                        uuid=uuids.inst3)}}),
            # host4 has one instance with os_distro=lin only
            ('host4', 'node4', {'instances': {uuids.inst2: objects.Instance(
                                                        uuid=uuids.inst4)}}),

        ]
        return [fakes.FakeHostState(host, node, values)
                for host, node, values in host_values]

    @mock.patch('nova.objects.InstanceList.fill_metadata')
    def test_multiplier_default(self, mock_fm):
        hostinfo_list = self._get_all_hosts()
        weighed_host = self._get_weighed_host(hostinfo_list)
        self.assertEqual(0.0, weighed_host.weight)
        # Nothing changes, we just returns the first in the list
        self.assertEqual('host1', weighed_host.obj.host)
        mock_fm.assert_not_called()

    @mock.patch('nova.objects.InstanceList.fill_metadata')
    def test_multiplier_windows_and_pc(self, mock_fm):
        self.flags(image_props_weight_multiplier=1.0, group='filter_scheduler')
        hostinfo_list = self._get_all_hosts()
        # we request for both windows and pc machine type
        weighed_host = self._get_weighed_host(
            hostinfo_list,
            request_spec=objects.RequestSpec(
                image=PROP_WIN_PC))
        self.assertEqual(1.0, weighed_host.weight)
        # only host2 has instances with both os_distro=windows and
        # hw_machine_type=pc
        self.assertEqual('host2', weighed_host.obj.host)
        mock_fm.assert_has_calls([mock.call(), mock.call(),
                                  mock.call(), mock.call()])

    @mock.patch('nova.objects.InstanceList.fill_metadata')
    def test_multiplier_pc(self, mock_fm):
        self.flags(image_props_weight_multiplier=1.0, group='filter_scheduler')
        hostinfo_list = self._get_all_hosts()
        weights = self.weight_handler.get_weighed_objects(
            self.weighers, hostinfo_list,
            weighing_properties=objects.RequestSpec(image=PROP_PC))
        # host2 and host3 have instances with pc machine type so are equally
        # weighed so we return the first of both.
        expected_weights = [{'weight': 1.0, 'host': 'host2'},
                            {'weight': 1.0, 'host': 'host3'},
                            {'weight': 0.0, 'host': 'host1'},
                            {'weight': 0.0, 'host': 'host4'}]
        self.assertEqual(expected_weights, [weigh.to_dict()
                                            for weigh in weights])
        self.assertEqual('host2', weights[0].obj.host)
        mock_fm.assert_has_calls([mock.call(), mock.call(),
                                  mock.call(), mock.call()])

    @mock.patch('nova.objects.InstanceList.fill_metadata')
    def test_multiplier_linux(self, mock_fm):
        self.flags(image_props_weight_multiplier=1.0, group='filter_scheduler')
        hostinfo_list = self._get_all_hosts()
        weights = self.weight_handler.get_weighed_objects(
            self.weighers, hostinfo_list,
            weighing_properties=objects.RequestSpec(image=PROP_LIN))
        # host3 and host4 have instances with linux distro but we favor
        # host3 given he has more instances having the same requested property
        expected_weights = [{'weight': 1.0, 'host': 'host3'},
                            {'weight': 0.5, 'host': 'host4'},
                            {'weight': 0.0, 'host': 'host1'},
                            {'weight': 0.0, 'host': 'host2'}]
        self.assertEqual(expected_weights, [weigh.to_dict()
                                            for weigh in weights])
        self.assertEqual('host3', weights[0].obj.host)
        mock_fm.assert_has_calls([mock.call(), mock.call(),
                                  mock.call(), mock.call()])

    @mock.patch('nova.objects.InstanceList.fill_metadata')
    def test_multiplier_per_property(self, mock_fm):
        self.flags(image_props_weight_multiplier=1.0, group='filter_scheduler')
        hostinfo_list = self._get_all_hosts()

        # For now, don't exclude any property check and boot with an image
        # using both hw_machine_type and os_distro properties.
        weights = self.weight_handler.get_weighed_objects(
            self.weighers, hostinfo_list,
            weighing_properties=objects.RequestSpec(image=PROP_LIN_PC))
        # host3 is preferred as it has both of the correct properties with
        # the right values.
        # host2 and host4 only support one of each property.
        expected_weights = [{'weight': 1.0, 'host': 'host3'},
                            {'weight': (1 / 3), 'host': 'host2'},
                            {'weight': (1 / 3), 'host': 'host4'},
                            {'weight': 0.0, 'host': 'host1'}]
        self.assertEqual(expected_weights, [weigh.to_dict()
                                            for weigh in weights])
        self.assertEqual('host3', weights[0].obj.host)

        # Now, let's exclude hw_machine_type property to be weighed.
        self.flags(image_props_weight_setting=['os_distro=1',
                                               'hw_machine_type=0'],
                   group='filter_scheduler')
        # Force a refresh of the settings since we updated them
        self.weighers[0]._parse_setting()
        weights = self.weight_handler.get_weighed_objects(
            self.weighers, hostinfo_list,
            weighing_properties=objects.RequestSpec(image=PROP_LIN_PC))
        # host3 and host4 have instances with linux distro but we favor
        # host3 given he has more instances having the same requested property
        expected_weights = [{'weight': 1.0, 'host': 'host3'},
                            {'weight': 0.5, 'host': 'host4'},
                            {'weight': 0.0, 'host': 'host1'},
                            {'weight': 0.0, 'host': 'host2'}]
        self.assertEqual(expected_weights, [weigh.to_dict()
                                            for weigh in weights])
        self.assertEqual('host3', weights[0].obj.host)
        mock_fm.assert_has_calls([mock.call(), mock.call(),
                                  mock.call(), mock.call()])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import fixtures
from unittest import mock

from lxml import etree
from oslo_utils.fixture import uuidsentinel
from oslo_utils import versionutils

from nova import conf
from nova import context as nova_context
from nova import objects
from nova.tests.functional.libvirt import base
from nova.virt.libvirt import driver as libvirt_driver
from nova.virt.libvirt import migration as libvirt_migration

CONF = conf.CONF


class LibvirtStatelessFirmwareTest(
    base.LibvirtMigrationMixin,
    base.ServersTestBase,
):

    # many move operations are admin-only
    ADMIN_API = True

    microversion = 'latest'

    def setUp(self):
        super().setUp()
        self.context = nova_context.get_admin_context()
        # Add the stateless firmware image to the glance fixture
        hw_fw_stateless_image = copy.deepcopy(self.glance.image1)
        hw_fw_stateless_image['id'] = uuidsentinel.fw_stateless_image_id
        hw_fw_stateless_image['properties']['hw_machine_type'] = 'q35'
        hw_fw_stateless_image['properties']['hw_firmware_type'] = 'uefi'
        hw_fw_stateless_image['properties']['hw_firmware_stateless'] = True
        self.glance.create(self.context, hw_fw_stateless_image)

        self._start_compute()

        self.guest_configs = {}
        orig_get_config = self.computes['compute1'].driver._get_guest_config

        def _get_guest_config(_self, *args, **kwargs):
            guest_config = orig_get_config(*args, **kwargs)
            instance = args[0]
            self.guest_configs[instance.uuid] = guest_config
            return self.guest_configs[instance.uuid]

        self.useFixture(fixtures.MonkeyPatch(
            'nova.virt.libvirt.LibvirtDriver._get_guest_config',
            _get_guest_config))

        # disk.rescue image_create ignoring privsep
        self.useFixture(fixtures.MockPatch(
            'nova.virt.libvirt.imagebackend._update_utime_ignore_eacces'))

        # dir to create 'unrescue.xml'
        def fake_path(_self, *args, **kwargs):
            return CONF.instances_path

        self.useFixture(fixtures.MonkeyPatch(
            'nova.virt.libvirt.utils.get_instance_path', fake_path))

    def _start_compute(self, hostname='compute1'):
        self.start_compute(
            hostname,
            libvirt_version=versionutils.convert_version_to_int(
                libvirt_driver.MIN_LIBVIRT_STATELESS_FIRMWARE
            ))
        caps = self.computes[hostname].driver.capabilities
        self.assertTrue(caps['supports_stateless_firmware'])

    def _create_server_with_stateless_firmware(self):
        server = self._create_server(
            image_uuid=uuidsentinel.fw_stateless_image_id,
            networks='none',
        )
        self.addCleanup(self._delete_server, server)
        return server

    def _create_server_without_stateless_firmware(self):
        server = self._create_server(
            image_uuid=self.glance.image1['id'],
            networks='none',
        )
        self.addCleanup(self._delete_server, server)
        return server

    def _assert_server_has_stateless_firmware(self, server_id):
        instance = objects.Instance.get_by_uuid(self.context, server_id)
        self.assertTrue(
            instance.image_meta.properties.hw_firmware_stateless
        )
        self.assertTrue(
            self.guest_configs[server_id].os_loader_stateless
        )
        del self.guest_configs[server_id]

    def _assert_server_has_no_stateless_firmware(self, server_id):
        instance = objects.Instance.get_by_uuid(self.context, server_id)
        self.assertIsNone(
            instance.image_meta.properties.get('hw_firmware_stateless')
        )
        self.assertIsNone(
            self.guest_configs[server_id].os_loader_stateless
        )
        del self.guest_configs[server_id]

    def test_create_server(self):
        """Assert new instance is created with stateless firmware only when
           the image has the required properties
        """
        server_with = self._create_server_with_stateless_firmware()
        self._assert_server_has_stateless_firmware(server_with['id'])

        server_without = self._create_server_without_stateless_firmware()
        self._assert_server_has_no_stateless_firmware(server_without['id'])

    def test_live_migrate_server(self):
        # create a server with stateless firmware
        self.server = self._create_server_with_stateless_firmware()
        self._assert_server_has_stateless_firmware(self.server['id'])

        self._start_compute('compute2')
        self.src = self.computes['compute1']
        self.dest = self.computes['compute2']

        self.migration_xml = None
        orig_get_updated_guest_xml = libvirt_migration.get_updated_guest_xml

        def migration_xml_wrapper(*args, **kwargs):
            self.migration_xml = orig_get_updated_guest_xml(*args, **kwargs)
            return self.migration_xml

        with mock.patch(
            'nova.virt.libvirt.migration.get_updated_guest_xml',
            side_effect=migration_xml_wrapper
        ) as fake_get_updated_guest_xml:
            self._live_migrate(self.server)

        fake_get_updated_guest_xml.assert_called_once()
        server = etree.fromstring(self.migration_xml)
        self.assertEqual('yes', server.find('./os/loader').get('stateless'))

    def test_migrate_server(self):
        self._start_compute('compute2')

        # create a server with stateless firmware
        server = self._create_server_with_stateless_firmware()
        self._assert_server_has_stateless_firmware(server['id'])

        # TODO(stephenfin): The mock of 'migrate_disk_and_power_off' should
        # probably be less...dumb
        with mock.patch(
            'nova.virt.libvirt.driver.LibvirtDriver'
            '.migrate_disk_and_power_off', return_value='{}',
        ):
            # cold migrate the server
            self._migrate_server(server)

        self._assert_server_has_stateless_firmware(server['id'])

        server = self._confirm_resize(server)

    def test_migrate_server_revert(self):
        self._start_compute('compute2')

        # create a server with stateless firmware
        server = self._create_server_with_stateless_firmware()
        self._assert_server_has_stateless_firmware(server['id'])

        # TODO(stephenfin): The mock of 'migrate_disk_and_power_off' should
        # probably be less...dumb
        with mock.patch(
            'nova.virt.libvirt.driver.LibvirtDriver'
            '.migrate_disk_and_power_off', return_value='{}',
        ):
            # cold migrate the server
            self._migrate_server(server)

        self._assert_server_has_stateless_firmware(server['id'])

        server = self._revert_resize(server)
        self._assert_server_has_stateless_firmware(server['id'])

    def test_shelve_and_unshelve_to_same_host(self):
        # create a server with stateless firmware
        server = self._create_server_with_stateless_firmware()
        self._assert_server_has_stateless_firmware(server['id'])

        # shelve the server
        server = self._shelve_server(server)

        # uneshelve the server. the server is started at the same compute
        server = self._unshelve_server(server)
        self._assert_server_has_stateless_firmware(server['id'])

    def test_shelve_and_unshelve_to_different_host(self):
        # create a server with stateless firmware
        server = self._create_server_with_stateless_firmware()
        self._assert_server_has_stateless_firmware(server['id'])

        # shelve the server
        server = self._shelve_server(server)

        # force down the compute node
        source_compute_id = self.admin_api.get_services(
            host='compute1', binary='nova-compute')[0]['id']
        self.computes['compute1'].stop()
        self.admin_api.put_service(
            source_compute_id, {'forced_down': 'true'})

        # start a new compute node and unshelve the server
        self._start_compute('compute2')
        server = self._unshelve_server(server)
        self.assertEqual(
            'compute2', server['OS-EXT-SRV-ATTR:hypervisor_hostname'])
        self._assert_server_has_stateless_firmware(server['id'])

    def test_evacuate_server(self):
        # create a server with stateless firmware
        server = self._create_server_with_stateless_firmware()
        self.assertEqual(
            'compute1', server['OS-EXT-SRV-ATTR:hypervisor_hostname'])
        self._assert_server_has_stateless_firmware(server['id'])

        # force down the compute node
        source_compute_id = self.admin_api.get_services(
            host='compute1', binary='nova-compute')[0]['id']
        self.computes['compute1'].stop()
        self.admin_api.put_service(
            source_compute_id, {'forced_down': 'true'})

        # start a new compute node and evecuate the server
        self._start_compute('compute2')
        server = self._evacuate_server(server, expected_host='compute2')
        self.assertEqual(
            'compute2', server['OS-EXT-SRV-ATTR:hypervisor_hostname'])
        self._assert_server_has_stateless_firmware(server['id'])

    def test_rescue_unrescue_server(self):
        # create a server without stateless firmware
        server = self._create_server_with_stateless_firmware()
        self._assert_server_has_stateless_firmware(server['id'])

        self.api.post_server_action(server['id'], {
            "rescue": {
                "rescue_image_ref": self.glance.image1['id']
            }
        })
        server = self._wait_for_state_change(server, 'RESCUE')

        # The instance object should still expect stateless firmware
        instance = objects.Instance.get_by_uuid(self.context, server['id'])
        self.assertTrue(
            instance.image_meta.properties.hw_firmware_stateless
        )

        # but the actual xml config should not
        self.assertIsNone(
            self.guest_configs[server['id']].os_loader_stateless
        )
        del self.guest_configs[server['id']]

        self.api.post_server_action(server['id'], {
            "unrescue": None
        })
        server = self._wait_for_state_change(server, 'ACTIVE')
        # NOTE(tkajinam): Unrescue restores the original xml file so skip
        #                 asserting its content.

    def test_rebuild_server(self):
        # create a server without stateless firmware
        server = self._create_server_without_stateless_firmware()
        self._assert_server_has_no_stateless_firmware(server['id'])

        # rebuild using the image with stateless firmware
        server = self._rebuild_server(
            server, uuidsentinel.fw_stateless_image_id)
        self._assert_server_has_stateless_firmware(server['id'])

        # rebuild using the image without stateless firmware
        server = self._rebuild_server(server, self.glance.image1['id'])
        self._assert_server_has_no_stateless_firmware(server['id'])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# TODO(stephenfin): Restrict the value to 'null' in a future API version
suspend = {
    'type': 'object',
    'properties': {
        'suspend': {},
    },
    'required': ['suspend'],
    'additionalProperties': False
}

# TODO(stephenfin): Restrict the value to 'null' in a future API version
resume = {
    'type': 'object',
    'properties': {
        'resume': {},
    },
    'required': ['resume'],
    'additionalProperties': False
}

suspend_response = {
    'type': 'null',
}

resume_response = {
    'type': 'null',
}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import fixtures
import nova
from oslo_config import cfg
from oslo_log import log as logging


CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class ManilaShare():
    def __init__(self, share_id, proto="NFS"):
        self.id = share_id
        self.size = 1
        self.availability_zone = "nova"
        self.created_at = "2015-09-18T10:25:24.000000"
        self.status = "available"
        self.name = "share_London"
        self.description = "My custom share London"
        self.project_id = "6a6a9c9eee154e9cb8cec487b98d36ab"
        self.snapshot_id = None
        self.share_network_id = "713df749-aac0-4a54-af52-10f6c991e80c"
        self.share_protocol = proto
        self.metadata = {"project": "my_app", "aim": "doc"}
        self.share_type = "25747776-08e5-494f-ab40-a64b9d20d8f7"
        self.volume_type = "default"
        self.is_public = True


class ManilaAccess():
    def __init__(self, access_type="ip"):
        self.access_level = "rw"
        self.state = "active"
        self.id = "507bf114-36f2-4f56-8cf4-857985ca87c1"
        if access_type == "ip":
            self.access_type = "ip"
            self.access_to = "192.168.0.1"
            self.access_key = None
        elif access_type == "cephx":
            self.access_type = "cephx"
            self.access_to = "nova"
            self.access_key = "mykey"


class ManilaFixture(fixtures.Fixture):
    """Fixture that mocks Manila APIs used by nova/share/manila.py"""

    def setUp(self):
        super().setUp()
        self.share_access = set()
        self.mock_get = self.useFixture(fixtures.MockPatch(
            'nova.share.manila.API.get',
            side_effect=self.fake_get)).mock
        self.mock_get_access = self.useFixture(fixtures.MockPatch(
            'nova.share.manila.API.get_access',
            side_effect=self.fake_get_access)).mock
        self.mock_allow = self.useFixture(fixtures.MockPatch(
            'nova.share.manila.API.allow',
            side_effect=self.fake_allow)).mock
        self.mock_deny = self.useFixture(fixtures.MockPatch(
            'nova.share.manila.API.deny',
            side_effect=self.fake_deny)).mock

    def fake_get(self, context, share_id):
        manila_share = ManilaShare(share_id)
        export_location = "10.0.0.50:/mnt/foo"
        return nova.share.manila.Share.from_manila_share(
            manila_share, export_location
        )

    def fake_get_share_status_error(self, context, share_id):
        manila_share = ManilaShare(share_id)
        manila_share.status = "error"
        export_location = "10.0.0.50:/mnt/foo"
        return nova.share.manila.Share.from_manila_share(
            manila_share, export_location
        )

    def fake_get_share_export_location_missing(self, context, share_id):
        manila_share = ManilaShare(share_id)
        export_location = None
        return nova.share.manila.Share.from_manila_share(
            manila_share, export_location
        )

    def fake_get_share_unknown_protocol(self, context, share_id):
        manila_share = ManilaShare(share_id)
        manila_share.share_protocol = "CIFS"
        export_location = "10.0.0.50:/mnt/foo"
        return nova.share.manila.Share.from_manila_share(
            manila_share, export_location
        )

    def fake_get_cephfs(self, context, share_id):
        manila_share = ManilaShare(share_id, "CEPHFS")
        export_location = "10.0.0.50:/mnt/foo"
        return nova.share.manila.Share.from_manila_share(
            manila_share, export_location
        )

    def fake_get_access(self, context, share_id, access_type, access_to):
        if share_id not in self.share_access:
            return None
        else:
            access = ManilaAccess()
            return nova.share.manila.Access.from_manila_access(access)

    def fake_get_access_cephfs(
        self, context, share_id, access_type, access_to
    ):
        access = ManilaAccess(access_type="cephx")
        return access

    def fake_allow(
        self, context, share_id, access_type, access_to, access_level
    ):
        self.share_access.add(share_id)
        self.fake_get_access(context, share_id, access_type, access_to)

    def fake_deny(self, context, share_id, access_type, access_to):
        self.share_access.discard(share_id)
        return 202



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import fixtures

from nova.tests.fixtures import libvirt as fakelibvirt
from nova.tests.functional.api import client
from nova.tests.functional.libvirt import base


class Bug2007968RegressionTest(base.ServersTestBase):
    """Regression test for bug 2007968
    """

    api_major_version = 'v2.1'
    microversion = 'latest'

    def setUp(self):
        super().setUp()

        self.useFixture(
            fixtures.MockPatch(
                'nova.virt.libvirt.driver.LibvirtDriver.'
                'migrate_disk_and_power_off',
                return_value='{}'
            )
        )

        self.flags(allow_resize_to_same_host=True)
        host_info = fakelibvirt.HostInfo()
        self.start_compute(host_info=host_info, hostname='compute1')

    def test_resize_with_smaller_memory_image_required(self):
        # Resize down from old_flavor to new_flavor.
        old_flavor = self._create_flavor(id='2007968_old', memory_mb=4096)
        new_flavor = self._create_flavor(id='2007968_new', memory_mb=512)

        # Now the instance for boot from image already has verification for
        # image min ram, but the instance for boot from volume does not have
        # such verification, so this bug only applies to instance that boot
        # from volume.
        server = self._create_server_boot_from_volume(
            networks='auto',
            image_args={'min_ram': 4096},
            flavor_id=old_flavor
        )

        # This can cause the instance for boot from volume to be allowed to
        # get to the resize verify status, but the instance's application runs
        # abnormally due to insufficient memory, and it may be killed by OOM.

        # After the fix, compute api will directly raise FlavorMemoryTooSmall
        # and will not continue the resize.
        ex = self.assertRaises(client.OpenStackApiException,
                               self._resize_server, server, new_flavor)
        self.assertEqual(400, ex.response.status_code)
        self.assertIn('Flavor\'s memory is too small for requested image.',
                      ex.response.text)



# Copyright 2010 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from webob import exc

from nova.api.openstack.compute.schemas import consoles as schema
from nova.api.openstack import wsgi
from nova.api import validation

_removal_reason = """\
This API only works with the Xen virt driver, which was deprecated in the
20.0.0 (Train) release.
It was removed in the 21.0.0 (Ussuri) release.
"""


class ConsolesController(wsgi.Controller):
    """(Removed) The Consoles controller for the OpenStack API.

    This was removed during the Ussuri release along with the nova-console
    service.
    """

    @wsgi.expected_errors(410)
    @wsgi.removed('21.0.0', _removal_reason)
    @validation.query_schema(schema.index_query)
    def index(self, req, server_id):
        raise exc.HTTPGone()

    @wsgi.expected_errors(410)
    @wsgi.removed('21.0.0', _removal_reason)
    @validation.schema(schema.create)
    def create(self, req, server_id, body):
        raise exc.HTTPGone()

    @wsgi.expected_errors(410)
    @wsgi.removed('21.0.0', _removal_reason)
    @validation.query_schema(schema.show_query)
    def show(self, req, server_id, id):
        raise exc.HTTPGone()

    @wsgi.expected_errors(410)
    @wsgi.removed('21.0.0', _removal_reason)
    def delete(self, req, server_id, id):
        raise exc.HTTPGone()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy

show_query = {
    'type': 'object',
    'properties': {},
    'additionalProperties': True,
}

show_query_v299 = copy.deepcopy(show_query)
show_query_v299['additionalProperties'] = False

show_response = {
    'type': 'object',
    'properties': {
        'console': {
            'type': 'object',
            'properties': {
                'instance_uuid': {'type': 'string', 'format': 'uuid'},
                'host': {'type': ['string', 'null']},
                'port': {'type': 'integer'},
                'internal_access_path': {
                    'oneOf': [
                        {
                            'type': 'null',
                        },
                        {
                            'type': 'string',
                            'format': 'uuid',
                        },
                    ],
                },
            },
            'required': [
                'instance_uuid', 'host', 'port', 'internal_access_path',
            ],
            'additionalProperties': False,
        },
    },
    'required': ['console'],
    'additionalProperties': False,
}

show_response_v299 = copy.deepcopy(show_response)
show_response_v299['properties']['console']['properties'].update({
    'tls_port': {'type': ['integer', 'null']},
})



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt
import fixtures
from oslo_serialization import jsonutils
from oslo_utils.fixture import uuidsentinel as uuids
import webob

from nova.api.openstack import api_version_request
from nova.api.openstack.compute import shelve as shelve_v21
from nova.compute import task_states
from nova.compute import vm_states
from nova import exception
from nova import test
from nova.tests.unit.api.openstack import fakes
from nova.tests.unit import fake_instance


@ddt.ddt
class ShelveControllerTest(test.NoDBTestCase):
    plugin = shelve_v21

    def setUp(self):
        super().setUp()
        self.controller = self.plugin.ShelveController()
        self.req = fakes.HTTPRequest.blank('')

    @ddt.data(
        exception.InstanceIsLocked(instance_uuid=uuids.instance),
        exception.UnexpectedTaskStateError(
            instance_uuid=uuids.instance, expected=None,
            actual=task_states.SHELVING),
    )
    @mock.patch('nova.compute.api.API.shelve')
    @mock.patch('nova.api.openstack.common.get_instance')
    def test_shelve__http_conflict_error(
        self, exc, mock_get_instance, mock_shelve,
    ):
        mock_get_instance.return_value = (
            fake_instance.fake_instance_obj(self.req.environ['nova.context']))
        mock_shelve.side_effect = exc

        self.assertRaises(
            webob.exc.HTTPConflict, self.controller._shelve,
            self.req, uuids.fake, body={'shelve': None})

    @ddt.data(
        exception.ForbiddenWithAccelerators(),
        exception.OperationNotSupportedForVTPM(
            instance_uuid=uuids.instance, operation='foo'),
        exception.OperationNotSupportedForVDPAInterface(
            instance_uuid=uuids.instance, operation='foo'),
    )
    @mock.patch('nova.compute.api.API.shelve')
    @mock.patch('nova.api.openstack.common.get_instance')
    def test_shelve_raise_badrequest_for_not_supported_feature(
        self, exc, mock_get_instance, mock_shelve,
    ):
        mock_get_instance.return_value = (
            fake_instance.fake_instance_obj(self.req.environ['nova.context']))
        mock_shelve.side_effect = exc
        self.assertRaises(
            webob.exc.HTTPBadRequest, self.controller._shelve,
            self.req, uuids.fake, body={'shelve': None})

    @mock.patch('nova.api.openstack.common.get_instance')
    def test_unshelve_locked_server(self, get_instance_mock):
        get_instance_mock.return_value = (
            fake_instance.fake_instance_obj(self.req.environ['nova.context']))
        self.stub_out('nova.compute.api.API.unshelve',
                      fakes.fake_actions_to_locked_server)
        self.assertRaises(webob.exc.HTTPConflict, self.controller._unshelve,
                          self.req, uuids.fake, body={'unshelve': None})

    @mock.patch('nova.api.openstack.common.get_instance')
    def test_shelve_offload_locked_server(self, get_instance_mock):
        get_instance_mock.return_value = (
            fake_instance.fake_instance_obj(self.req.environ['nova.context']))
        self.stub_out('nova.compute.api.API.shelve_offload',
                      fakes.fake_actions_to_locked_server)
        self.assertRaises(webob.exc.HTTPConflict,
                          self.controller._shelve_offload,
                          self.req, uuids.fake, body={'shelveOffload': None})


class UnshelveServerControllerTestV277(test.NoDBTestCase):
    """Server controller test for microversion 2.77

    Add availability_zone parameter to unshelve a shelved-offloaded server of
    2.77 microversion.
    """
    wsgi_api_version = '2.77'

    def setUp(self):
        super(UnshelveServerControllerTestV277, self).setUp()
        self.mock_neutron_extension_list = self.useFixture(
            fixtures.MockPatch(
                'nova.network.neutron.API._refresh_neutron_extensions_cache'
            )
        ).mock
        self.mock_neutron_extension_list.return_value = {'extensions': []}
        self.controller = shelve_v21.ShelveController()
        self.req = fakes.HTTPRequest.blank(
                '/%s/servers/a/action' % fakes.FAKE_PROJECT_ID,
                use_admin_context=True, version=self.wsgi_api_version)

    def fake_get_instance(self):
        ctxt = self.req.environ['nova.context']
        return fake_instance.fake_instance_obj(
            ctxt, uuid=fakes.FAKE_UUID, vm_state=vm_states.SHELVED_OFFLOADED)

    @mock.patch('nova.api.openstack.common.get_instance')
    def test_unshelve_with_az_pre_2_77_failed(self, mock_get_instance):
        """Make sure specifying an AZ before microversion 2.77 is ignored."""
        instance = self.fake_get_instance()
        mock_get_instance.return_value = instance

        body = {
            'unshelve': {
                'availability_zone': 'us-east'
            }}
        self.req.body = jsonutils.dump_as_bytes(body)
        self.req.api_version_request = (
            api_version_request.APIVersionRequest('2.76')
        )
        with mock.patch.object(
            self.controller.compute_api, 'unshelve'
        ) as mock_unshelve:
            self.controller._unshelve(self.req, fakes.FAKE_UUID, body=body)
        mock_unshelve.assert_called_once_with(
            self.req.environ['nova.context'],
            instance,
        )

    @mock.patch('nova.compute.api.API.unshelve')
    @mock.patch('nova.api.openstack.common.get_instance')
    def test_unshelve_with_none_pre_2_77_success(
            self, mock_get_instance, mock_unshelve):
        """Make sure we can unshelve server with None
        before microversion 2.77.
        """
        instance = self.fake_get_instance()
        mock_get_instance.return_value = instance

        body = {'unshelve': None}
        self.req.body = jsonutils.dump_as_bytes(body)
        self.req.api_version_request = (api_version_request.
                APIVersionRequest('2.76'))
        self.controller._unshelve(self.req, fakes.FAKE_UUID, body=body)
        mock_unshelve.assert_called_once_with(
            self.req.environ['nova.context'],
            instance,
        )

    @mock.patch('nova.compute.api.API.unshelve')
    @mock.patch('nova.api.openstack.common.get_instance')
    def test_unshelve_with_empty_dict_with_v2_77_failed(
            self, mock_get_instance, mock_unshelve):
        """Make sure we cannot unshelve server with empty dict."""
        instance = self.fake_get_instance()
        mock_get_instance.return_value = instance

        body = {'unshelve': {}}
        self.req.body = jsonutils.dump_as_bytes(body)
        exc = self.assertRaises(exception.ValidationError,
                                self.controller._unshelve,
                                self.req, fakes.FAKE_UUID,
                                body=body)
        self.assertIn("\'availability_zone\' is a required property", str(exc))

    def test_invalid_az_name_with_int(self):
        body = {
            'unshelve': {
                'availability_zone': 1234
            }}
        self.req.body = jsonutils.dump_as_bytes(body)
        self.assertRaises(exception.ValidationError,
                          self.controller._unshelve,
                          self.req, fakes.FAKE_UUID,
                          body=body)

    def test_no_az_value(self):
        body = {
            'unshelve': {
                'availability_zone': None
            }}
        self.req.body = jsonutils.dump_as_bytes(body)
        self.assertRaises(
                exception.ValidationError,
                self.controller._unshelve,
                self.req,
                fakes.FAKE_UUID, body=body)

    def test_unshelve_with_additional_param(self):
        body = {
            'unshelve': {
                'availability_zone': 'us-east',
                'additional_param': 1
            }}
        self.req.body = jsonutils.dump_as_bytes(body)
        exc = self.assertRaises(
            exception.ValidationError,
            self.controller._unshelve, self.req,
            fakes.FAKE_UUID, body=body)
        self.assertIn("Additional properties are not allowed", str(exc))


class UnshelveServerControllerTestV291(test.NoDBTestCase):
    """Server controller test for microversion 2.91

    Add host parameter to unshelve a shelved-offloaded server of
    2.91 microversion.
    """
    wsgi_api_version = '2.91'

    def setUp(self):
        super(UnshelveServerControllerTestV291, self).setUp()
        self.controller = shelve_v21.ShelveController()
        self.req = fakes.HTTPRequest.blank(
                '/%s/servers/a/action' % fakes.FAKE_PROJECT_ID,
                use_admin_context=True, version=self.wsgi_api_version)

    def fake_get_instance(self):
        ctxt = self.req.environ['nova.context']
        return fake_instance.fake_instance_obj(
            ctxt, uuid=fakes.FAKE_UUID, vm_state=vm_states.SHELVED_OFFLOADED)

    @mock.patch('nova.api.openstack.common.get_instance')
    def test_unshelve_with_az_pre_2_91(self, mock_get_instance):
        """Make sure specifying an AZ before microversion 2.91
        is still working.
        """
        instance = self.fake_get_instance()
        mock_get_instance.return_value = instance

        body = {
            'unshelve': {
                'availability_zone': 'us-east',
            }}
        self.req.body = jsonutils.dump_as_bytes(body)
        self.req.api_version_request = (
                api_version_request.APIVersionRequest('2.77'))
        with mock.patch.object(
            self.controller.compute_api, 'unshelve'
        ) as mock_unshelve:
            self.controller._unshelve(self.req, fakes.FAKE_UUID, body=body)
        mock_unshelve.assert_called_once_with(
            self.req.environ['nova.context'],
            instance,
            new_az='us-east',
        )

    @mock.patch('nova.api.openstack.common.get_instance')
    def test_unshelve_without_parameters_2_91(self, mock_get_instance):
        """Make sure not specifying parameters with microversion 2.91
        is working.
        """
        instance = self.fake_get_instance()
        mock_get_instance.return_value = instance

        body = {
            'unshelve': None
        }
        self.req.body = jsonutils.dump_as_bytes(body)
        self.req.api_version_request = (
                api_version_request.APIVersionRequest('2.91'))
        with mock.patch.object(
                self.controller.compute_api, 'unshelve') as mock_unshelve:
            self.controller._unshelve(self.req, fakes.FAKE_UUID, body=body)
        mock_unshelve.assert_called_once_with(
            self.req.environ['nova.context'],
            instance,
        )

    @mock.patch('nova.api.openstack.common.get_instance')
    def test_unshelve_with_az_2_91(self, mock_get_instance):
        """Make sure specifying an AZ with microversion 2.91
        is working.
        """
        instance = self.fake_get_instance()
        mock_get_instance.return_value = instance

        body = {
            'unshelve': {
                'availability_zone': 'us-east',
            }}
        self.req.body = jsonutils.dump_as_bytes(body)
        self.req.api_version_request = (
                api_version_request.APIVersionRequest('2.91'))
        with mock.patch.object(
                self.controller.compute_api, 'unshelve') as mock_unshelve:
            self.controller._unshelve(self.req, fakes.FAKE_UUID, body=body)
        mock_unshelve.assert_called_once_with(
            self.req.environ['nova.context'],
            instance,
            new_az='us-east',
            host=None,
        )

    @mock.patch('nova.api.openstack.common.get_instance')
    def test_unshelve_with_az_none_2_91(self, mock_get_instance):
        """Make sure specifying an AZ to none (unpin server)
        is working.
        """
        instance = self.fake_get_instance()
        mock_get_instance.return_value = instance

        body = {
            'unshelve': {
                'availability_zone': None,
            }}
        self.req.body = jsonutils.dump_as_bytes(body)
        self.req.api_version_request = (
                api_version_request.APIVersionRequest('2.91'))
        with mock.patch.object(
                self.controller.compute_api, 'unshelve') as mock_unshelve:
            self.controller._unshelve(self.req, fakes.FAKE_UUID, body=body)
        mock_unshelve.assert_called_once_with(
            self.req.environ['nova.context'],
            instance,
            new_az=None,
            host=None,
        )

    @mock.patch('nova.api.openstack.common.get_instance')
    def test_unshelve_with_host_2_91(self, mock_get_instance):
        """Make sure specifying a host with microversion 2.91
        is working.
        """
        instance = self.fake_get_instance()
        mock_get_instance.return_value = instance

        body = {
            'unshelve': {
                'host': 'server02',
            }}
        self.req.body = jsonutils.dump_as_bytes(body)
        self.req.api_version_request = (
                api_version_request.APIVersionRequest('2.91'))
        with mock.patch.object(
                self.controller.compute_api, 'unshelve') as mock_unshelve:
            self.controller._unshelve(self.req, fakes.FAKE_UUID, body=body)
        mock_unshelve.assert_called_once_with(
            self.req.environ['nova.context'],
            instance,
            host='server02',
        )

    @mock.patch('nova.compute.api.API.unshelve')
    @mock.patch('nova.api.openstack.common.get_instance')
    def test_unshelve_with_az_and_host_with_v2_91(
            self, mock_get_instance, mock_unshelve):
        """Make sure specifying a host and an availability_zone with
        microversion 2.91 is working.
        """
        instance = self.fake_get_instance()
        mock_get_instance.return_value = instance

        body = {
            'unshelve': {
                'availability_zone': 'us-east',
                'host': 'server01',
            }}
        self.req.body = jsonutils.dump_as_bytes(body)
        self.req.api_version_request = (
                api_version_request.APIVersionRequest('2.91'))

        self.controller._unshelve(self.req, fakes.FAKE_UUID, body=body)
        self.controller.compute_api.unshelve.assert_called_once_with(
            self.req.environ['nova.context'],
            instance,
            new_az='us-east',
            host='server01',
        )

    def test_invalid_az_name_with_int(self):
        body = {
            'unshelve': {
                'host': 1234
            }}
        self.req.body = jsonutils.dump_as_bytes(body)
        self.assertRaises(
                exception.ValidationError,
                self.controller._unshelve,
                self.req,
                fakes.FAKE_UUID,
                body=body)

    def test_no_az_value(self):
        body = {
            'unshelve': {
                'host': None
            }}
        self.req.body = jsonutils.dump_as_bytes(body)
        self.assertRaises(
                exception.ValidationError,
                self.controller._unshelve,
                self.req,
                fakes.FAKE_UUID, body=body)

    def test_invalid_host_fqdn_with_int(self):
        body = {
            'unshelve': {
                'host': 1234
            }}
        self.req.body = jsonutils.dump_as_bytes(body)
        self.assertRaises(
                exception.ValidationError,
                self.controller._unshelve,
                self.req,
                fakes.FAKE_UUID,
                body=body)

    def test_no_host(self):
        body = {
            'unshelve': {
                'host': None
            }}
        self.req.body = jsonutils.dump_as_bytes(body)
        self.assertRaises(exception.ValidationError,
                          self.controller._unshelve,
                          self.req, fakes.FAKE_UUID,
                          body=body)

    def test_unshelve_with_additional_param(self):
        body = {
            'unshelve': {
                'host': 'server01',
                'additional_param': 1
            }}
        self.req.body = jsonutils.dump_as_bytes(body)
        exc = self.assertRaises(
            exception.ValidationError,
            self.controller._unshelve, self.req,
            fakes.FAKE_UUID, body=body)
        self.assertIn("Invalid input for field/attribute unshelve.", str(exc))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_policy import policy

from nova.policies import base


BASE_POLICY_NAME = 'compute:server:topology:%s'

server_topology_policies = [
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'index',
        check_str=base.PROJECT_READER_OR_ADMIN,
        description="Show the NUMA topology data for a server",
        operations=[
            {
                'method': 'GET',
                'path': '/servers/{server_id}/topology'
            }
        ],
        scope_types=['project']),
    policy.DocumentedRuleDefault(
        # Control host NUMA node and cpu pinning information
        name=BASE_POLICY_NAME % 'host:index',
        check_str=base.ADMIN,
        description="Show the NUMA topology data for a server with host "
        "NUMA ID and CPU pinning information",
        operations=[
            {
                'method': 'GET',
                'path': '/servers/{server_id}/topology'
            }
        ],
        scope_types=['project']),
]


def list_rules():
    return server_topology_policies



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from nova.tests.fixtures import libvirt as fakelibvirt
from nova.tests.functional.libvirt import test_vgpu


class VGPUTestVolumeOPs(test_vgpu.VGPUTestBase):

    FAKE_LIBVIRT_VERSION = 8000000

    def setUp(self):
        super().setUp()
        extra_spec = {"resources:VGPU": "1"}
        self.flavor = self._create_flavor(extra_spec=extra_spec)

        # Start compute1 supporting only nvidia-11
        self.flags(
            enabled_mdev_types=fakelibvirt.NVIDIA_11_VGPU_TYPE,
            group='devices')

        self.compute1 = self.start_compute_with_vgpu('host1')

    def test_create_vgpu_server_volume_attach(self):

        # Create a single instance against a specific compute node.
        server = self._create_server(
            image_uuid='155d900f-4e14-4e4c-a73d-069cbf4541e6',
            flavor_id=self.flavor, host=self.compute1.host,
            networks='auto', expected_state='ACTIVE')
        server_id = server['id']

        self.assert_mdev_usage(self.compute1, expected_amount=1)

        # Attach a volume to the instance
        # reuse IMAGE_BACKED_VOL for simplicity
        self.api.post_server_volume(
            server_id,
            {
                'volumeAttachment': {
                    'volumeId': self.cinder.IMAGE_BACKED_VOL
                }
            }
        )
        self._wait_for_volume_attach(server_id, self.cinder.IMAGE_BACKED_VOL)
        return server_id

    def test_create_vgpu_server_volume_detach(self):
        server_id = self.test_create_vgpu_server_volume_attach()
        # Detach the volume from the instance
        # DELETE /servers/{server_id}/os-volume_attachments/{volume_id} is
        # async but as we are using CastAsCall it's sync in our func tests
        self.api.delete_server_volume(
            server_id, self.cinder.IMAGE_BACKED_VOL)
        self._wait_for_volume_detach(
            server_id, self.cinder.IMAGE_BACKED_VOL)



# Copyright (c) 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import collections
from unittest import mock

from oslo_config import cfg
from oslo_serialization import jsonutils
from oslo_utils.fixture import uuidsentinel as uuids

from nova import exception
from nova import objects
from nova.objects import fields
from nova.pci.request import PCI_REMOTE_MANAGED_TAG
from nova.pci import stats
from nova.pci import whitelist
from nova import test
from nova.tests.unit.pci import fakes

CONF = cfg.CONF
fake_pci_1 = {
    'compute_node_id': 1,
    'address': '0000:00:00.1',
    'product_id': 'p1',
    'vendor_id': 'v1',
    'status': 'available',
    'extra_k1': 'v1',
    'request_id': None,
    'numa_node': 0,
    'dev_type': fields.PciDeviceType.STANDARD,
    'parent_addr': None,
    }


fake_pci_2 = dict(fake_pci_1, vendor_id='v2',
                  product_id='p2',
                  address='0000:00:00.2',
                  numa_node=1)


fake_pci_3 = dict(fake_pci_1, address='0000:00:00.3')

fake_pci_4 = dict(fake_pci_1, vendor_id='v3',
                  product_id='p3',
                  address='0000:00:00.3',
                  numa_node= None)

pci_requests = [objects.InstancePCIRequest(count=1,
                    spec=[{'vendor_id': 'v1'}]),
                objects.InstancePCIRequest(count=1,
                    spec=[{'vendor_id': 'v2'}])]


pci_requests_multiple = [objects.InstancePCIRequest(count=1,
                             spec=[{'vendor_id': 'v1'}]),
                         objects.InstancePCIRequest(count=3,
                          spec=[{'vendor_id': 'v2'}])]


class PciDeviceStatsTestCase(test.NoDBTestCase):

    @staticmethod
    def _get_fake_requests(vendor_ids=None, numa_policy=None, count=1):
        if not vendor_ids:
            vendor_ids = ['v1', 'v2']

        specs = [{'vendor_id': vendor_id} for vendor_id in vendor_ids]

        return [objects.InstancePCIRequest(count=count, spec=[spec],
                numa_policy=numa_policy) for spec in specs]

    def _create_fake_devs(self):
        self.fake_dev_1 = objects.PciDevice.create(None, fake_pci_1)
        self.fake_dev_2 = objects.PciDevice.create(None, fake_pci_2)
        self.fake_dev_3 = objects.PciDevice.create(None, fake_pci_3)
        self.fake_dev_4 = objects.PciDevice.create(None, fake_pci_4)

        for dev in [self.fake_dev_1, self.fake_dev_2,
                    self.fake_dev_3, self.fake_dev_4]:
            self.pci_stats.add_device(dev)

    def _add_fake_devs_with_numa(self):
        fake_pci = dict(fake_pci_1, vendor_id='v4', product_id='pr4')
        devs = [dict(fake_pci, product_id='pr0', numa_node=0, ),
                dict(fake_pci, product_id='pr1', numa_node=1),
                dict(fake_pci, product_id='pr_none', numa_node=None)]

        for dev in devs:
            self.pci_stats.add_device(objects.PciDevice.create(None, dev))

    def setUp(self):
        super(PciDeviceStatsTestCase, self).setUp()
        self.pci_stats = stats.PciDeviceStats(objects.NUMATopology())
        # The following two calls need to be made before adding the devices.
        patcher = fakes.fake_pci_whitelist()
        self.addCleanup(patcher.stop)
        self._create_fake_devs()

    def test_add_device(self):
        self.assertEqual(len(self.pci_stats.pools), 4)
        self.assertEqual(set([d['vendor_id'] for d in self.pci_stats]),
                         set(['v1', 'v2', 'v3']))
        self.assertEqual([d['count'] for d in self.pci_stats], [1, 1, 1, 1])

    def test_remove_device(self):
        self.assertEqual(len(self.pci_stats.pools), 4)
        self.pci_stats.remove_device(self.fake_dev_2)
        self.assertEqual(len(self.pci_stats.pools), 3)
        self.assertEqual(self.pci_stats.pools[0]['count'], 1)
        self.assertEqual(self.pci_stats.pools[0]['vendor_id'], 'v1')
        self.assertEqual(self.pci_stats.pools[1]['count'], 1)
        self.assertEqual(self.pci_stats.pools[1]['vendor_id'], 'v1')

    def test_remove_device_exception(self):
        self.pci_stats.remove_device(self.fake_dev_2)
        self.assertRaises(exception.PciDevicePoolEmpty,
                          self.pci_stats.remove_device,
                          self.fake_dev_2)

    def test_pci_stats_equivalent(self):
        pci_stats2 = stats.PciDeviceStats(objects.NUMATopology())
        for dev in [self.fake_dev_1,
                    self.fake_dev_2,
                    self.fake_dev_3,
                    self.fake_dev_4]:
            pci_stats2.add_device(dev)
        self.assertEqual(self.pci_stats, pci_stats2)

    def test_pci_stats_not_equivalent(self):
        pci_stats2 = stats.PciDeviceStats(objects.NUMATopology())
        for dev in [self.fake_dev_1,
                    self.fake_dev_2,
                    self.fake_dev_3]:
            pci_stats2.add_device(dev)
        self.assertNotEqual(self.pci_stats, pci_stats2)

    def test_object_create(self):
        m = self.pci_stats.to_device_pools_obj()
        new_stats = stats.PciDeviceStats(objects.NUMATopology(), m)

        self.assertEqual(len(new_stats.pools), 4)
        self.assertEqual([d['count'] for d in new_stats], [1, 1, 1, 1])
        self.assertEqual(set([d['vendor_id'] for d in new_stats]),
                         set(['v1', 'v2', 'v3']))

    def test_apply_requests(self):
        self.assertEqual(len(self.pci_stats.pools), 4)
        self.pci_stats.apply_requests(pci_requests, {})
        self.assertEqual(len(self.pci_stats.pools), 2)
        self.assertEqual(self.pci_stats.pools[0]['vendor_id'], 'v1')
        self.assertEqual(self.pci_stats.pools[0]['count'], 1)

    def test_apply_requests_failed(self):
        self.assertRaises(
            exception.PciDeviceRequestFailed,
            self.pci_stats.apply_requests,
            pci_requests_multiple,
            {},
        )

    def test_support_requests(self):
        self.assertTrue(self.pci_stats.support_requests(pci_requests, {}))
        self.assertEqual(len(self.pci_stats.pools), 4)
        self.assertEqual([d['count'] for d in self.pci_stats], [1, 1, 1, 1])

    def test_support_requests_failed(self):
        self.assertFalse(
            self.pci_stats.support_requests(pci_requests_multiple, {}))
        self.assertEqual(len(self.pci_stats.pools), 4)
        self.assertEqual([d['count'] for d in self.pci_stats], [1, 1, 1, 1])

    def test_support_requests_numa(self):
        cells = [
            objects.InstanceNUMACell(
                id=0, cpuset=set(), pcpuset=set(), memory=0),
            objects.InstanceNUMACell(
                id=1, cpuset=set(), pcpuset=set(), memory=0),
        ]
        self.assertTrue(
            self.pci_stats.support_requests(pci_requests, {}, cells)
        )

    def test_support_requests_numa_failed(self):
        cells = [
            objects.InstanceNUMACell(
                id=0, cpuset=set(), pcpuset=set(), memory=0),
        ]
        self.assertFalse(
            self.pci_stats.support_requests(pci_requests, {}, cells)
        )

    def test_support_requests_no_numa_info(self):
        cells = [
            objects.InstanceNUMACell(
                id=0, cpuset=set(), pcpuset=set(), memory=0),
        ]
        pci_requests = self._get_fake_requests(vendor_ids=['v3'])
        self.assertTrue(
            self.pci_stats.support_requests(pci_requests, {}, cells)
        )

        # 'legacy' is the default numa_policy so the result must be same
        pci_requests = self._get_fake_requests(vendor_ids=['v3'],
            numa_policy = fields.PCINUMAAffinityPolicy.LEGACY)
        self.assertTrue(
            self.pci_stats.support_requests(pci_requests, {}, cells)
        )

    def test_support_requests_numa_pci_numa_policy_preferred(self):
        # numa node 0 has 2 devices with vendor_id 'v1'
        # numa node 1 has 1 device with vendor_id 'v2'
        # we request two devices with vendor_id 'v1' and 'v2'.
        # pci_numa_policy is 'preferred' so we can ignore numa affinity
        cells = [
            objects.InstanceNUMACell(
                id=0, cpuset=set(), pcpuset=set(), memory=0),
        ]
        pci_requests = self._get_fake_requests(
            numa_policy=fields.PCINUMAAffinityPolicy.PREFERRED)

        self.assertTrue(
            self.pci_stats.support_requests(pci_requests, {}, cells)
        )

    def test_support_requests_no_numa_info_pci_numa_policy_required(self):
        # pci device with vendor_id 'v3' has numa_node=None.
        # pci_numa_policy is 'required' so we can't use this device
        cells = [
            objects.InstanceNUMACell(
                id=0, cpuset=set(), pcpuset=set(), memory=0),
        ]
        pci_requests = self._get_fake_requests(vendor_ids=['v3'],
            numa_policy=fields.PCINUMAAffinityPolicy.REQUIRED)

        self.assertFalse(
            self.pci_stats.support_requests(pci_requests, {}, cells)
        )

    def test_filter_pools_for_socket_affinity_no_socket(self):
        self.pci_stats.numa_topology = objects.NUMATopology(
                cells=[objects.NUMACell(socket=None)])

        self.assertEqual(
            [],
            self.pci_stats._filter_pools_for_socket_affinity(
                self.pci_stats.pools, [objects.InstanceNUMACell()]))

    def test_filter_pools_for_socket_affinity(self):
        self.pci_stats.numa_topology = objects.NUMATopology(
                cells=[objects.NUMACell(id=1, socket=1)])

        pools = self.pci_stats._filter_pools_for_socket_affinity(
            self.pci_stats.pools, [objects.InstanceNUMACell(id=1)])
        self.assertEqual(1, len(pools))
        self.assertEqual('p2', pools[0]['product_id'])
        self.assertEqual('v2', pools[0]['vendor_id'])

    def test_consume_requests(self):
        devs = self.pci_stats.consume_requests(pci_requests)
        self.assertEqual(2, len(devs))
        self.assertEqual(set(['v1', 'v2']),
                         set([dev.vendor_id for dev in devs]))

    def test_consume_requests_empty(self):
        devs = self.pci_stats.consume_requests([])
        self.assertEqual(0, len(devs))

    def test_consume_requests_failed(self):
        self.assertRaises(
            exception.PciDeviceRequestFailed,
            self.pci_stats.consume_requests,
            pci_requests_multiple,
        )

    def test_consume_requests_numa(self):
        cells = [
            objects.InstanceNUMACell(
                id=0, cpuset=set(), pcpuset=set(), memory=0),
            objects.InstanceNUMACell(
                id=1, cpuset=set(), pcpuset=set(), memory=0),
        ]
        devs = self.pci_stats.consume_requests(pci_requests, cells)
        self.assertEqual(2, len(devs))
        self.assertEqual(set(['v1', 'v2']),
                         set([dev.vendor_id for dev in devs]))

    def test_consume_requests_numa_failed(self):
        cells = [
            objects.InstanceNUMACell(
                id=0, cpuset=set(), pcpuset=set(), memory=0),
        ]
        self.assertRaises(
            exception.PciDeviceRequestFailed,
            self.pci_stats.consume_requests,
            pci_requests,
            cells,
        )

    def test_consume_requests_no_numa_info(self):
        cells = [
            objects.InstanceNUMACell(
                id=0, cpuset=set(), pcpuset=set(), memory=0),
        ]
        pci_request = [objects.InstancePCIRequest(count=1,
                                                  spec=[{'vendor_id': 'v3'}])]
        devs = self.pci_stats.consume_requests(pci_request, cells)
        self.assertEqual(1, len(devs))
        self.assertEqual(set(['v3']),
                         set([dev.vendor_id for dev in devs]))

    def _test_consume_requests_numa_policy(self, cell_ids, policy,
            expected, vendor_id='v4', count=1):
        """Base test for 'consume_requests' function.

        Create three devices with vendor_id of 'v4': 'pr0' in NUMA node 0,
        'pr1' in NUMA node 1, 'pr_none' without NUMA affinity info. Attempt to
        consume a PCI request with a single device with ``vendor_id`` using the
        provided ``cell_ids`` and ``policy``. Compare result against
        ``expected``.
        """
        self._add_fake_devs_with_numa()
        cells = [
            objects.InstanceNUMACell(
                id=id, cpuset=set(), pcpuset=set(), memory=0)
            for id in cell_ids]

        pci_requests = self._get_fake_requests(vendor_ids=[vendor_id],
            numa_policy=policy, count=count)

        if expected is None:
            self.assertRaises(
                exception.PciDeviceRequestFailed,
                self.pci_stats.consume_requests,
                pci_requests,
                cells,
            )
        else:
            devs = self.pci_stats.consume_requests(pci_requests, cells)
            self.assertEqual(set(expected),
                             set([dev.product_id for dev in devs]))

    def test_consume_requests_numa_policy_required(self):
        """Ensure REQUIRED policy will ensure NUMA affinity.

        Policy is 'required' which means we must use a device with strict NUMA
        affinity. Request a device from NUMA node 0, which contains such a
        device, and ensure it's used.
        """
        self._test_consume_requests_numa_policy(
            [0], fields.PCINUMAAffinityPolicy.REQUIRED, ['pr0'])

    def test_consume_requests_numa_policy_required_fail(self):
        """Ensure REQUIRED policy will *only* provide NUMA affinity.

        Policy is 'required' which means we must use a device with strict NUMA
        affinity. Request a device from NUMA node 999, which does not contain
        any suitable devices, and ensure nothing is returned.
        """
        self._test_consume_requests_numa_policy(
            [999], fields.PCINUMAAffinityPolicy.REQUIRED, None)

    def test_consume_requests_numa_policy_legacy(self):
        """Ensure LEGACY policy will ensure NUMA affinity if possible.

        Policy is 'legacy' which means we must use a device with strict NUMA
        affinity or no provided NUMA affinity. Request a device from NUMA node
        0, which contains such a device, and ensure it's used.
        """
        self._test_consume_requests_numa_policy(
            [0], fields.PCINUMAAffinityPolicy.LEGACY, ['pr0'])

    def test_consume_requests_numa_policy_legacy_fallback(self):
        """Ensure LEGACY policy will fallback to no NUMA affinity.

        Policy is 'legacy' which means we must use a device with strict NUMA
        affinity or no provided NUMA affinity. Request a device from NUMA node
        999, which contains no such device, and ensure we fallback to the
        device without any NUMA affinity.
        """
        self._test_consume_requests_numa_policy(
            [999], fields.PCINUMAAffinityPolicy.LEGACY, ['pr_none'])

    def test_consume_requests_numa_policy_legacy_multiple(self):
        """Ensure LEGACY policy will use best policy for multiple devices.

        Policy is 'legacy' which means we must use a device with strict NUMA
        affinity or no provided NUMA affinity. Request two devices from NUMA
        node 0, which contains only one such device, and ensure we use that
        device and the next best thing for the second device.
        """
        self._test_consume_requests_numa_policy(
            [0], fields.PCINUMAAffinityPolicy.PREFERRED, ['pr0', 'pr_none'],
            count=2)

    def test_consume_requests_numa_policy_legacy_fail(self):
        """Ensure REQUIRED policy will *not* provide NUMA non-affinity.

        Policy is 'legacy' which means we must use a device with strict NUMA
        affinity or no provided NUMA affinity. Request a device with
        ``vendor_id`` of ``v2``, which can only be found in NUMA node 1, from
        NUMA node 0, and ensure nothing is returned.
        """
        self._test_consume_requests_numa_policy(
            [0], fields.PCINUMAAffinityPolicy.LEGACY, None, vendor_id='v2')

    def test_consume_requests_numa_policy_preferred(self):
        """Ensure PREFERRED policy will ensure NUMA affinity if possible.

        Policy is 'preferred' which means we must use a device with any level
        of NUMA affinity. Request a device from NUMA node 0, which contains
        an affined device, and ensure it's used.
        """
        self._test_consume_requests_numa_policy(
            [0], fields.PCINUMAAffinityPolicy.PREFERRED, ['pr0'])

    def test_consume_requests_numa_policy_preferred_fallback_a(self):
        """Ensure PREFERRED policy will fallback to no NUMA affinity.

        Policy is 'preferred' which means we must use a device with any level
        of NUMA affinity. Request a device from NUMA node 999, which contains
        no such device, and ensure we fallback to the device without any NUMA
        affinity.
        """
        self._test_consume_requests_numa_policy(
            [999], fields.PCINUMAAffinityPolicy.PREFERRED, ['pr_none'])

    def test_consume_requests_numa_policy_preferred_fallback_b(self):
        """Ensure PREFERRED policy will fallback to different NUMA affinity.

        Policy is 'preferred' which means we must use a device with any level
        of NUMA affinity. Request a device with ``vendor_id`` of ``v2``, which
        can only be found in NUMA node 1, from NUMA node 0, and ensure we
        fallback to this device.
        """
        self._test_consume_requests_numa_policy(
            [0], fields.PCINUMAAffinityPolicy.PREFERRED, ['p2'],
            vendor_id='v2')

    def test_consume_requests_numa_policy_preferred_multiple_a(self):
        """Ensure PREFERRED policy will use best policy for multiple devices.

        Policy is 'preferred' which means we must use a device with any level
        of NUMA affinity. Request two devices from NUMA node 0, which contains
        only one such device, and ensure we use that device and gracefully
        degrade for the other device.
        """
        self._test_consume_requests_numa_policy(
            [0], fields.PCINUMAAffinityPolicy.PREFERRED, ['pr0', 'pr_none'],
            count=2)

    def test_consume_requests_numa_policy_preferred_multiple_b(self):
        """Ensure PREFERRED policy will use best policy for multiple devices.

        Policy is 'preferred' which means we must use a device with any level
        of NUMA affinity. Request three devices from NUMA node 0, which
        contains only one such device, and ensure we use that device and
        gracefully degrade for the other devices.
        """
        self._test_consume_requests_numa_policy(
            [0], fields.PCINUMAAffinityPolicy.PREFERRED,
            ['pr0', 'pr_none', 'pr1'], count=3)

    @mock.patch(
        'nova.pci.whitelist.Whitelist._parse_white_list_from_config')
    def test_device_spec_parsing(self, mock_whitelist_parse):
        device_spec = {"product_id": "0001", "vendor_id": "8086"}
        CONF.set_override('device_spec', jsonutils.dumps(device_spec), 'pci')
        pci_stats = stats.PciDeviceStats(objects.NUMATopology())
        pci_stats.add_device(self.fake_dev_2)
        pci_stats.remove_device(self.fake_dev_2)
        self.assertEqual(1, mock_whitelist_parse.call_count)


class PciDeviceStatsWithTagsTestCase(test.NoDBTestCase):

    def setUp(self):
        super(PciDeviceStatsWithTagsTestCase, self).setUp()
        device_spec = [
            jsonutils.dumps(
                {
                    "vendor_id": "1137",
                    "product_id": "0071",
                    "address": "*:0a:00.*",
                    "physical_network": "physnet1",
                }
            ),
            jsonutils.dumps({"vendor_id": "1137", "product_id": "0072"}),
            jsonutils.dumps(
                {
                    "vendor_id": "15b3",
                    "product_id": "101e",
                    "remote_managed": "true",
                }
            ),
            jsonutils.dumps({"vendor_id": "15b3", "product_id": "101c"}),
            jsonutils.dumps(
                {
                    "vendor_id": "15b3",
                    "product_id": "1018",
                    "remote_managed": "false",
                }
            ),
            jsonutils.dumps(
                {
                    "vendor_id": "15b4",
                    "product_id": "102e",
                    "live_migratable": "true",
                }
            ),
            jsonutils.dumps(
                {
                    "vendor_id": "15b5",
                    "product_id": "102f",
                    "live_migratable": "false",
                }
            ),
        ]
        self.flags(device_spec=device_spec, group="pci")
        dev_filter = whitelist.Whitelist(device_spec)
        self.pci_stats = stats.PciDeviceStats(
            objects.NUMATopology(),
            dev_filter=dev_filter)

    def _create_pci_devices(self):
        self.pci_tagged_devices = []
        for dev in range(4):
            pci_dev = {'compute_node_id': 1,
                       'address': '0000:0a:00.%d' % dev,
                       'vendor_id': '1137',
                       'product_id': '0071',
                       'status': 'available',
                       'request_id': None,
                       'dev_type': 'type-PCI',
                       'parent_addr': None,
                       'numa_node': 0}
            self.pci_tagged_devices.append(objects.PciDevice.create(None,
                                                                    pci_dev))

        self.pci_untagged_devices = []
        for dev in range(3):
            pci_dev = {'compute_node_id': 1,
                       'address': '0000:0b:00.%d' % dev,
                       'vendor_id': '1137',
                       'product_id': '0072',
                       'status': 'available',
                       'request_id': None,
                       'dev_type': 'type-PCI',
                       'parent_addr': None,
                       'numa_node': 0}
            self.pci_untagged_devices.append(objects.PciDevice.create(None,
                                                                      pci_dev))

        self.locally_managed_netdevs = []
        self.remote_managed_netdevs = []
        self.remote_managed_netdevs.append(
            objects.PciDevice.create(
                None, {
                    'compute_node_id': 1,
                    'address': '0000:0c:00.1',
                    'vendor_id': '15b3',
                    'product_id': '101e',
                    'status': 'available',
                    'request_id': None,
                    'dev_type': fields.PciDeviceType.SRIOV_VF,
                    'parent_addr': '0000:0c:00.0',
                    'numa_node': 0,
                    "capabilities": {"vpd": {
                        "card_serial_number": "MT2113X00000"}}
                }))

        # For testing implicit remote_managed == False tagging.
        self.locally_managed_netdevs.append(
            objects.PciDevice.create(
                None, {
                    'compute_node_id': 1,
                    'address': '0000:0d:00.1',
                    'vendor_id': '15b3',
                    'product_id': '101c',
                    'status': 'available',
                    'request_id': None,
                    'dev_type': fields.PciDeviceType.SRIOV_VF,
                    'parent_addr': '0000:0d:00.0',
                    'numa_node': 0}))

        # For testing explicit remote_managed == False tagging.
        self.locally_managed_netdevs.append(
            objects.PciDevice.create(
                None, {
                    'compute_node_id': 1,
                    'address': '0000:0e:00.1',
                    'vendor_id': '15b3',
                    'product_id': '101c',
                    'status': 'available',
                    'request_id': None,
                    'dev_type': fields.PciDeviceType.SRIOV_VF,
                    'parent_addr': '0000:0e:00.0',
                    'numa_node': 0}))

        # live_migratable set to true
        # Warning: 'extra_info' field is set by the init or create method
        self.live_migratable = []
        self.live_migratable.append(
            objects.PciDevice.create(
                None, {
                    'compute_node_id': 1,
                    'address': '0000:0f:00.1',
                    'vendor_id': '15b4',
                    'product_id': '102e',
                    'status': 'available',
                    'request_id': None,
                    'dev_type': fields.PciDeviceType.SRIOV_VF,
                    'parent_addr': '0000:0f:00.0',
                    'numa_node': 0,
                }))

        # live_migratable set to false
        # Warning: 'extra_info' field is set by the init or create method
        self.non_live_migratable = []
        self.non_live_migratable.append(
            objects.PciDevice.create(
                None, {
                    'compute_node_id': 1,
                    'address': '0000:10:00.1',
                    'vendor_id': '15b5',
                    'product_id': '102f',
                    'status': 'available',
                    'request_id': None,
                    'dev_type': fields.PciDeviceType.SRIOV_VF,
                    'parent_addr': '0000:10:00.0',
                    'numa_node': 0,
                }))

        for dev in self.pci_tagged_devices:
            self.pci_stats.add_device(dev)

        for dev in self.pci_untagged_devices:
            self.pci_stats.add_device(dev)

        for dev in self.remote_managed_netdevs:
            self.pci_stats.add_device(dev)

        for dev in self.locally_managed_netdevs:
            self.pci_stats.add_device(dev)

        for dev in self.live_migratable:
            self.pci_stats.add_device(dev)

        for dev in self.non_live_migratable:
            self.pci_stats.add_device(dev)

    def _assertPoolContent(self, pool, vendor_id, product_id, count, **tags):
        self.assertEqual(vendor_id, pool['vendor_id'])
        self.assertEqual(product_id, pool['product_id'])
        self.assertEqual(count, pool['count'])
        if tags:
            for k, v in tags.items():
                self.assertEqual(v, pool[k])

    def _assertPools(self):
        nr_tagged = len(self.pci_tagged_devices)
        nr_untagged = len(self.pci_untagged_devices)
        nr_remote = len(self.remote_managed_netdevs)
        nr_local = len(self.locally_managed_netdevs)
        nr_lm = len(self.live_migratable)
        nr_nlm = len(self.non_live_migratable)
        self.assertEqual(
            nr_tagged + nr_untagged + nr_remote + nr_local + nr_lm + nr_nlm,
            len(self.pci_stats.pools),
        )
        # Pools are ordered based on the number of keys. 'product_id',
        # 'vendor_id' are always part of the keys. When tags are present,
        # they are also part of the keys.

        # 3 pools for the pci_untagged_devices
        devs = []
        j = 0
        for i in range(j, j + nr_untagged):
            self._assertPoolContent(self.pci_stats.pools[i], '1137', '0072', 1)
            devs += self.pci_stats.pools[i]['devices']
        self.assertEqual(self.pci_untagged_devices, devs)
        j += nr_untagged

        # 4 pools for the pci_tagged_devices'
        devs = []
        for i in range(j, j + nr_tagged):
            self._assertPoolContent(
                self.pci_stats.pools[i],
                "1137",
                "0071",
                1,
                physical_network="physnet1",
            )
            devs += self.pci_stats.pools[i]['devices']
        self.assertEqual(self.pci_tagged_devices, devs)
        j += nr_tagged

        # one with remote_managed_netdevs
        devs = []
        for i in range(j, j + nr_remote):
            self._assertPoolContent(
                self.pci_stats.pools[i],
                "15b3",
                "101e",
                1,
                remote_managed="true",
            )
            devs += self.pci_stats.pools[i]['devices']
        self.assertEqual(self.remote_managed_netdevs, devs)
        j += nr_remote

        # two with locally_managed_netdevs
        devs = []
        for i in range(j, j + nr_local):
            self._assertPoolContent(
                self.pci_stats.pools[i],
                "15b3",
                "101c",
                1,
                remote_managed="false",
            )
            devs += self.pci_stats.pools[i]['devices']
        self.assertEqual(self.locally_managed_netdevs, devs)
        j += nr_local

        # one live_migratable
        devs = []
        for i in range(j, j + nr_lm):
            self._assertPoolContent(
                self.pci_stats.pools[i],
                "15b4",
                "102e",
                1,
            )
            devs += self.pci_stats.pools[i]['devices']
        self.assertEqual(self.live_migratable, devs)
        j += nr_lm

        # one non_live_migratable
        devs = []
        for i in range(j, j + nr_nlm):
            self._assertPoolContent(
                self.pci_stats.pools[i],
                "15b5",
                "102f",
                1,
            )
            devs += self.pci_stats.pools[i]['devices']
        self.assertEqual(self.non_live_migratable, devs)
        j += nr_nlm

    def test_add_devices(self):
        self._create_pci_devices()
        self._assertPools()

    def test_consume_requests(self):
        self._create_pci_devices()
        pci_requests = [objects.InstancePCIRequest(count=1,
                            spec=[{'physical_network': 'physnet1'}]),
                        objects.InstancePCIRequest(count=1,
                            spec=[{'vendor_id': '1137',
                                   'product_id': '0072'}]),
                        objects.InstancePCIRequest(count=1,
                            spec=[{'vendor_id': '15b3',
                                   'product_id': '101e',
                                   PCI_REMOTE_MANAGED_TAG: 'True'}]),
                        objects.InstancePCIRequest(count=1,
                            spec=[{'vendor_id': '15b3',
                                   'product_id': '101c',
                                   PCI_REMOTE_MANAGED_TAG: 'False'}]),
                        objects.InstancePCIRequest(count=1,
                            spec=[{'vendor_id': '15b3',
                                   'product_id': '101c',
                                   PCI_REMOTE_MANAGED_TAG: 'False'}]),
                        objects.InstancePCIRequest(count=1,
                            spec=[{'vendor_id': '15b4',
                                   'product_id': '102e',
                                   'live_migratable': 'true'}]),
                        objects.InstancePCIRequest(count=1,
                            spec=[{'vendor_id': '15b5',
                                   'product_id': '102f',
                                   'live_migratable': 'false'}]),
                        ]
        devs = self.pci_stats.consume_requests(pci_requests)
        self.assertEqual(7, len(devs))
        self.assertEqual(set(['0071', '0072', '101e', '101c', '102e', '102f']),
                         set([dev.product_id for dev in devs]))
        self._assertPoolContent(self.pci_stats.pools[0], '1137', '0072', 0)
        self._assertPoolContent(self.pci_stats.pools[1], '1137', '0072', 1)
        self._assertPoolContent(self.pci_stats.pools[2], '1137', '0072', 1)

        self._assertPoolContent(self.pci_stats.pools[3], '1137', '0071', 0,
                                physical_network='physnet1')
        self._assertPoolContent(self.pci_stats.pools[4], '1137', '0071', 1,
                                physical_network='physnet1')
        self._assertPoolContent(self.pci_stats.pools[5], '1137', '0071', 1,
                                physical_network='physnet1')
        self._assertPoolContent(self.pci_stats.pools[6], '1137', '0071', 1,
                                physical_network='physnet1')

        self._assertPoolContent(self.pci_stats.pools[7], '15b3', '101e', 0,
                                remote_managed='true')
        self._assertPoolContent(self.pci_stats.pools[8], '15b3', '101c', 0,
                                remote_managed='false')
        self._assertPoolContent(self.pci_stats.pools[9], '15b3', '101c', 0,
                                remote_managed='false')
        self._assertPoolContent(self.pci_stats.pools[10], '15b4', '102e', 0,
                                live_migratable='true')
        self._assertPoolContent(self.pci_stats.pools[11], '15b5', '102f', 0,
                                live_migratable='false')

    def test_add_device_no_devspec(self):
        self._create_pci_devices()
        pci_dev = {'compute_node_id': 1,
                   'address': '0000:0c:00.1',
                   'vendor_id': '2345',
                   'product_id': '0172',
                   'status': 'available',
                   'parent_addr': None,
                   'request_id': None}
        pci_dev_obj = objects.PciDevice.create(None, pci_dev)
        self.pci_stats.add_device(pci_dev_obj)
        # There should be no change
        self.assertIsNone(
            self.pci_stats._create_pool_keys_from_dev(pci_dev_obj))
        self._assertPools()

    def test_remove_device_no_devspec(self):
        self._create_pci_devices()
        pci_dev = {'compute_node_id': 1,
                   'address': '0000:0c:00.1',
                   'vendor_id': '2345',
                   'product_id': '0172',
                   'status': 'available',
                   'parent_addr': None,
                   'request_id': None}
        pci_dev_obj = objects.PciDevice.create(None, pci_dev)
        self.pci_stats.remove_device(pci_dev_obj)
        # There should be no change
        self.assertIsNone(
            self.pci_stats._create_pool_keys_from_dev(pci_dev_obj))
        self._assertPools()

    def test_remove_device(self):
        self._create_pci_devices()
        dev1 = self.pci_untagged_devices.pop()
        self.pci_stats.remove_device(dev1)
        dev2 = self.pci_tagged_devices.pop()
        self.pci_stats.remove_device(dev2)
        self._assertPools()

    def test_update_device_splits_the_pool(self):
        # Update device type of one of the device from type-VF to
        # type-PF. Verify if the existing pool is updated and a new
        # pool is created with dev_type type-PF.
        vfs = []
        for i in range(3):
            dev = objects.PciDevice(
                    compute_node_id=1,
                    address="0000:0a:00.%d" % i,
                    vendor_id="1137",
                    product_id="0071",
                    status="available",
                    dev_type="type-VF",
                    parent_addr="0000:0a:01.0",
                    numa_node=0
                )
            vfs.append(dev)
            self.pci_stats.add_device(dev)

        self.assertEqual(1, len(self.pci_stats.pools))
        self.assertEqual(3, self.pci_stats.pools[0]["count"])
        self.assertEqual(vfs, self.pci_stats.pools[0]["devices"])

        dev = vfs.pop()
        dev.dev_type = 'type-PF'
        dev.parent_addr = None
        self.pci_stats.update_device(dev)
        self.assertEqual(2, len(self.pci_stats.pools))
        self.assertEqual(2, self.pci_stats.pools[0]["count"])
        self.assertEqual(vfs, self.pci_stats.pools[0]["devices"])
        self.assertEqual(1, self.pci_stats.pools[1]["count"])
        self.assertEqual([dev], self.pci_stats.pools[1]["devices"])

    def test_only_vfs_from_the_same_parent_are_pooled(self):
        pf1_vfs = []
        for i in range(2):
            dev = objects.PciDevice(
                    compute_node_id=1,
                    address="0000:0a:00.%d" % i,
                    vendor_id="15b3",
                    product_id="1018",
                    status="available",
                    dev_type="type-VF",
                    parent_addr="0000:0a:01.0",
                    numa_node=0
                )
            pf1_vfs.append(dev)
            self.pci_stats.add_device(dev)

        pf2_vfs = []
        for i in range(2):
            dev = objects.PciDevice(
                    compute_node_id=1,
                    address="0000:0b:00.%d" % i,
                    vendor_id="15b3",
                    product_id="1018",
                    status="available",
                    dev_type="type-VF",
                    parent_addr="0000:0b:01.0",
                    numa_node=0
                )
            pf2_vfs.append(dev)
            self.pci_stats.add_device(dev)

        self.assertEqual(2, len(self.pci_stats.pools))
        self.assertEqual(2, self.pci_stats.pools[0]["count"])
        self.assertEqual(pf1_vfs, self.pci_stats.pools[0]["devices"])
        self.assertEqual(2, len(self.pci_stats.pools))
        self.assertEqual(2, self.pci_stats.pools[1]["count"])
        self.assertEqual(pf2_vfs, self.pci_stats.pools[1]["devices"])


class PciDeviceStatsPlacementSupportTestCase(test.NoDBTestCase):

    def test_device_spec_rc_and_traits_ignored_during_pooling(self):
        """Assert that resource_class and traits from the device spec are not
        used as discriminator for pool creation.
        """
        device_spec = [
            jsonutils.dumps(
                {
                    "resource_class": "foo",
                    "address": "*:81:00.1",
                    "traits": "gold",
                }
            ),
            jsonutils.dumps(
                {
                    "resource_class": "baar",
                    "address": "*:81:00.2",
                    "traits": "silver",
                }
            ),
        ]
        self.flags(device_spec=device_spec, group="pci")
        dev_filter = whitelist.Whitelist(device_spec)
        pci_stats = stats.PciDeviceStats(
            objects.NUMATopology(),
            dev_filter=dev_filter)
        pci_dev1 = objects.PciDevice(
            vendor_id="dead",
            product_id="beef",
            address="0000:81:00.1",
            parent_addr="0000:81:00.0",
            numa_node=0,
            dev_type="type-VF",
        )
        pci_dev2 = objects.PciDevice(
            vendor_id="dead",
            product_id="beef",
            address="0000:81:00.2",
            parent_addr="0000:81:00.0",
            numa_node=0,
            dev_type="type-VF",
        )
        # the two device matched by different device_specs with different
        # resource_class and traits fields
        pci_stats.add_device(pci_dev1)
        pci_stats.add_device(pci_dev2)

        # but they are put in the same pool as all the other fields are
        # matching
        self.assertEqual(1, len(pci_stats.pools))
        self.assertEqual(2, pci_stats.pools[0]["count"])

    def test_filter_pools_for_spec_ignores_rc_and_traits_in_spec(self):
        """Assert that resource_class and traits are ignored in the pci
        request spec during matching the request to pools.
        """
        pci_stats = stats.PciDeviceStats(objects.NUMATopology())
        pools = [{"vendor_id": "dead", "product_id": "beef"}]

        matching_pools = pci_stats._filter_pools_for_spec(
            pools=pools,
            request=objects.InstancePCIRequest(
                spec=[
                    {
                        "vendor_id": "dead",
                        "product_id": "beef",
                        "resource_class": "foo",
                        "traits": "blue",
                    }
                ]
            ),
        )

        self.assertEqual(pools, matching_pools)

    def test_populate_pools_metadata_from_assigned_devices(self):
        device_spec = [
            jsonutils.dumps(
                {
                    "address": "0000:81:00.*",
                }
            ),
        ]
        self.flags(device_spec=device_spec, group="pci")
        dev_filter = whitelist.Whitelist(device_spec)
        pci_stats = stats.PciDeviceStats(
            objects.NUMATopology(),
            dev_filter=dev_filter)
        pci_dev1 = objects.PciDevice(
            vendor_id="dead",
            product_id="beef",
            address="0000:81:00.1",
            parent_addr="0000:81:00.0",
            numa_node=0,
            dev_type="type-VF",
        )
        pci_dev2 = objects.PciDevice(
            vendor_id="dead",
            product_id="beef",
            address="0000:81:00.2",
            parent_addr="0000:81:00.0",
            numa_node=0,
            dev_type="type-VF",
        )
        pci_stats.add_device(pci_dev1)
        pci_dev1.extra_info = {'rp_uuid': uuids.rp1}
        pci_stats.add_device(pci_dev2)
        pci_dev2.extra_info = {'rp_uuid': uuids.rp1}

        self.assertEqual(1, len(pci_stats.pools))

        pci_stats.populate_pools_metadata_from_assigned_devices()

        self.assertEqual(uuids.rp1, pci_stats.pools[0]['rp_uuid'])

    def test_populate_pools_metadata_from_assigned_devices_device_without_rp(
        self
    ):
        device_spec = [
            jsonutils.dumps(
                {
                    "address": "0000:81:00.*",
                }
            ),
        ]
        self.flags(device_spec=device_spec, group="pci")
        dev_filter = whitelist.Whitelist(device_spec)
        pci_stats = stats.PciDeviceStats(
            objects.NUMATopology(),
            dev_filter=dev_filter)
        pci_dev1 = objects.PciDevice(
            vendor_id="dead",
            product_id="beef",
            address="0000:81:00.1",
            parent_addr="0000:81:00.0",
            numa_node=0,
            dev_type="type-VF",
        )
        pci_stats.add_device(pci_dev1)

        self.assertEqual(1, len(pci_stats.pools))

        pci_stats.populate_pools_metadata_from_assigned_devices()

        self.assertNotIn('rp_uuid', pci_stats.pools[0])

    def test_populate_pools_metadata_from_assigned_devices_multiple_rp(self):
        device_spec = [
            jsonutils.dumps(
                {
                    "address": "0000:81:00.*",
                }
            ),
        ]
        self.flags(device_spec=device_spec, group="pci")
        dev_filter = whitelist.Whitelist(device_spec)
        pci_stats = stats.PciDeviceStats(
            objects.NUMATopology(),
            dev_filter=dev_filter)
        pci_dev1 = objects.PciDevice(
            compute_node_id=1,
            vendor_id="dead",
            product_id="beef",
            address="0000:81:00.1",
            parent_addr="0000:81:00.0",
            numa_node=0,
            dev_type="type-VF",
        )
        pci_dev2 = objects.PciDevice(
            compute_node_id=1,
            vendor_id="dead",
            product_id="beef",
            address="0000:81:00.2",
            parent_addr="0000:81:00.0",
            numa_node=0,
            dev_type="type-VF",
        )
        pci_stats.add_device(pci_dev1)
        pci_dev1.extra_info = {'rp_uuid': uuids.rp1}
        pci_stats.add_device(pci_dev2)
        pci_dev2.extra_info = {'rp_uuid': uuids.rp2}

        self.assertEqual(1, len(pci_stats.pools))

        self.assertRaises(
            ValueError,
            pci_stats.populate_pools_metadata_from_assigned_devices,
        )


class PciDeviceStatsProviderMappingTestCase(test.NoDBTestCase):
    def setUp(self):
        super().setUp()
        # for simplicity accept any devices
        device_spec = [
            jsonutils.dumps(
                {
                    "address": "*:*:*.*",
                }
            ),
        ]
        self.flags(device_spec=device_spec, group="pci")
        self.dev_filter = whitelist.Whitelist(device_spec)
        self.pci_stats = stats.PciDeviceStats(
            objects.NUMATopology(), dev_filter=self.dev_filter
        )
        # add devices represented by different RPs in placement
        # two VFs on the same PF
        self.vf1 = objects.PciDevice(
            compute_node_id=1,
            vendor_id="dead",
            product_id="beef",
            address="0000:81:00.1",
            parent_addr="0000:81:00.0",
            numa_node=0,
            dev_type="type-VF",
        )
        self.vf2 = objects.PciDevice(
            compute_node_id=1,
            vendor_id="dead",
            product_id="beef",
            address="0000:81:00.2",
            parent_addr="0000:81:00.0",
            numa_node=0,
            dev_type="type-VF",
        )
        self.pci_stats.add_device(self.vf1)
        self.vf1.extra_info = {'rp_uuid': uuids.pf1}
        self.pci_stats.add_device(self.vf2)
        self.vf2.extra_info = {'rp_uuid': uuids.pf1}
        # two PFs pf2 and pf3 (pf1 is used for the paren of the above VFs)
        self.pf2 = objects.PciDevice(
            compute_node_id=1,
            vendor_id="dead",
            product_id="beef",
            address="0000:82:00.0",
            parent_addr=None,
            numa_node=0,
            dev_type="type-PF",
        )
        self.pci_stats.add_device(self.pf2)
        self.pf2.extra_info = {'rp_uuid': uuids.pf2}

        self.pf3 = objects.PciDevice(
            compute_node_id=1,
            vendor_id="dead",
            product_id="beef",
            address="0000:83:00.0",
            parent_addr=None,
            numa_node=0,
            dev_type="type-PF",
        )
        self.pci_stats.add_device(self.pf3)
        self.pf3.extra_info = {'rp_uuid': uuids.pf3}
        # a PCI
        self.pci1 = objects.PciDevice(
            compute_node_id=1,
            vendor_id="dead",
            product_id="beef",
            address="0000:84:00.0",
            parent_addr=None,
            numa_node=0,
            dev_type="type-PCI",
        )
        self.pci_stats.add_device(self.pci1)
        self.pci1.extra_info = {'rp_uuid': uuids.pci1}

        # populate the RP -> pool mapping from the devices to its pools
        self.pci_stats.populate_pools_metadata_from_assigned_devices()

        # we have 1 pool for the two VFs then the rest has it own pool one by
        # one
        self.num_pools = 4
        self.assertEqual(self.num_pools, len(self.pci_stats.pools))
        self.num_devs = 5
        self.assertEqual(
            self.num_devs, sum(pool["count"] for pool in self.pci_stats.pools)
        )

    def test_support_request_unrestricted(self):
        reqs = []
        for dev_type in ["type-VF", "type-PF", "type-PCI"]:
            req = objects.InstancePCIRequest(
                count=1,
                alias_name='a-dev',
                spec=[
                    {
                        "vendor_id": "dead",
                        "product_id": "beef",
                        "dev_type": dev_type,
                    }
                ],
            )
            reqs.append(req)

        # an empty mapping means unrestricted by any provider
        # we have devs for all type so each request should fit
        self.assertTrue(self.pci_stats.support_requests(reqs, {}))

        # the support_requests call is expected not to consume any device
        self.assertEqual(self.num_pools, len(self.pci_stats.pools))
        self.assertEqual(
            self.num_devs, sum(pool["count"] for pool in self.pci_stats.pools)
        )

        # now apply the same request to consume the pools
        self.pci_stats.apply_requests(reqs, {})
        # we have consumed a 3 devs (a VF, a PF, and a PCI)
        self.assertEqual(
            self.num_devs - 3,
            sum(pool["count"] for pool in self.pci_stats.pools),
        )
        # the empty pools are purged. We have one pool for the remaining VF
        # and the remaining PF
        self.assertEqual(2, len(self.pci_stats.pools))

    def test_support_request_restricted_by_provider_mapping(self):
        pf_req = objects.InstancePCIRequest(
            count=1,
            alias_name='a-dev',
            request_id=uuids.req1,
            spec=[
                {
                    "vendor_id": "dead",
                    "product_id": "beef",
                    "dev_type": "type-PF",
                }
            ],
        )

        # simulate the placement restricted the possible RPs to pf3
        self.assertTrue(
            self.pci_stats.support_requests(
                [pf_req], {f"{uuids.req1}-0": [uuids.pf3]}
            )
        )

        # the support_requests call is expected not to consume any device
        self.assertEqual(self.num_pools, len(self.pci_stats.pools))
        self.assertEqual(
            self.num_devs, sum(pool["count"] for pool in self.pci_stats.pools)
        )

        # now apply the request and see if the right device is consumed
        self.pci_stats.apply_requests(
            [pf_req], {f"{uuids.req1}-0": [uuids.pf3]}
        )

        self.assertEqual(self.num_pools - 1, len(self.pci_stats.pools))
        self.assertEqual(
            self.num_devs - 1,
            sum(pool["count"] for pool in self.pci_stats.pools),
        )
        # pf3 is not available in the pools any more
        self.assertEqual(
            {uuids.pf1, uuids.pf2, uuids.pci1},
            {pool['rp_uuid'] for pool in self.pci_stats.pools},
        )

    def test_support_request_restricted_by_provider_mapping_does_not_fit(self):
        pf_req = objects.InstancePCIRequest(
            count=1,
            alias_name='a-dev',
            request_id=uuids.req1,
            spec=[
                {
                    "vendor_id": "dead",
                    "product_id": "beef",
                    "dev_type": "type-PF",
                }
            ],
        )

        # Simulate that placement returned an allocation candidate with a PF
        # that is not in the pools anymore, e.g. filtered out by numa cell.
        # We expect the request to fail
        self.assertFalse(
            self.pci_stats.support_requests(
                [pf_req], {f"{uuids.req1}-0": [uuids.pf4]}
            )
        )
        self.assertRaises(
            exception.PciDeviceRequestFailed,
            self.pci_stats.apply_requests,
            [pf_req],
            {f"{uuids.req1}-0": [uuids.pf4]},
        )
        # and the pools are not changed
        self.assertEqual(self.num_pools, len(self.pci_stats.pools))
        self.assertEqual(
            self.num_devs, sum(pool["count"] for pool in self.pci_stats.pools)
        )

    def test_support_request_neutron_port_based_request_ignore_mapping(self):
        # by not having the alias_name set this becomes a neutron port based
        # PCI request
        pf_req = objects.InstancePCIRequest(
            count=1,
            request_id=uuids.req1,
            spec=[
                {
                    "vendor_id": "dead",
                    "product_id": "beef",
                    "dev_type": "type-PF",
                }
            ],
        )

        # Simulate that placement returned an allocation candidate with a PF
        # that is not in the pools anymore, e.g. filtered out by numa cell.
        # We expect that the placement selection is ignored for neutron port
        # based requests so this request should fit as we have PFs in the pools
        self.assertTrue(
            self.pci_stats.support_requests(
                [pf_req], {f"{uuids.req1}-0": [uuids.pf4]}
            )
        )
        self.pci_stats.apply_requests(
            [pf_req],
            {f"{uuids.req1}-0": [uuids.pf4]},
        )
        # and a PF is consumed
        self.assertEqual(self.num_pools - 1, len(self.pci_stats.pools))
        self.assertEqual(
            self.num_devs - 1,
            sum(pool["count"] for pool in self.pci_stats.pools),
        )

    def test_support_request_req_with_count_2(self):
        # now ask for two PFs in a single request
        pf_req = objects.InstancePCIRequest(
            count=2,
            alias_name='a-dev',
            request_id=uuids.req1,
            spec=[
                {
                    "vendor_id": "dead",
                    "product_id": "beef",
                    "dev_type": "type-PF",
                }
            ],
        )

        # Simulate that placement returned one candidate RP for both PF reqs
        mapping = {
            f"{uuids.req1}-0": [uuids.pf2],
            f"{uuids.req1}-1": [uuids.pf3],
        }
        # so the request fits
        self.assertTrue(self.pci_stats.support_requests([pf_req], mapping))
        self.pci_stats.apply_requests([pf_req], mapping)
        # and both PFs are consumed
        self.assertEqual(self.num_pools - 2, len(self.pci_stats.pools))
        self.assertEqual(
            self.num_devs - 2,
            sum(pool["count"] for pool in self.pci_stats.pools),
        )
        self.assertEqual(
            {uuids.pf1, uuids.pci1},
            {pool['rp_uuid'] for pool in self.pci_stats.pools},
        )

    def test_support_requests_multiple_reqs(self):
        # request both a VF and a PF
        vf_req = objects.InstancePCIRequest(
            count=1,
            alias_name='a-dev',
            request_id=uuids.vf_req,
            spec=[
                {
                    "vendor_id": "dead",
                    "product_id": "beef",
                    "dev_type": "type-VF",
                }
            ],
        )
        pf_req = objects.InstancePCIRequest(
            count=1,
            alias_name='a-dev',
            request_id=uuids.pf_req,
            spec=[
                {
                    "vendor_id": "dead",
                    "product_id": "beef",
                    "dev_type": "type-PF",
                }
            ],
        )

        # Simulate that placement returned one candidate RP for both reqs
        mapping = {
            # the VF is represented by the parent PF RP
            f"{uuids.vf_req}-0": [uuids.pf1],
            f"{uuids.pf_req}-0": [uuids.pf3],
        }
        # so the request fits
        self.assertTrue(
            self.pci_stats.support_requests([vf_req, pf_req], mapping)
        )
        self.pci_stats.apply_requests([vf_req, pf_req], mapping)
        # and the proper devices are consumed
        # Note that the VF pool still has a device so it remains
        self.assertEqual(self.num_pools - 1, len(self.pci_stats.pools))
        self.assertEqual(
            self.num_devs - 2,
            sum(pool["count"] for pool in self.pci_stats.pools),
        )
        self.assertEqual(
            {uuids.pf1, uuids.pf2, uuids.pci1},
            {pool['rp_uuid'] for pool in self.pci_stats.pools},
        )

    def test_apply_gets_requested_uuids_from_pci_req(self):
        pf_req = objects.InstancePCIRequest(
            count=1,
            alias_name='a-dev',
            request_id=uuids.req1,
            spec=[
                {
                    "vendor_id": "dead",
                    "product_id": "beef",
                    "dev_type": "type-PF",
                    # Simulate that the scheduler already allocate a candidate
                    # and the mapping is stored in the request.
                    # The allocation restricts that we can only consume from
                    # PF3
                    "rp_uuids": ",".join([uuids.pf3])
                }
            ],
        )

        # call apply with None mapping signalling that the allocation is
        # already done and the resulted mapping is stored in the request
        self.pci_stats.apply_requests([pf_req], provider_mapping=None)

        # assert that the right device is consumed
        self.assertEqual(self.num_pools - 1, len(self.pci_stats.pools))
        self.assertEqual(
            self.num_devs - 1,
            sum(pool["count"] for pool in self.pci_stats.pools),
        )
        # pf3 is not available in the pools anymore
        self.assertEqual(
            {uuids.pf1, uuids.pf2, uuids.pci1},
            {pool['rp_uuid'] for pool in self.pci_stats.pools},
        )

    def _create_two_pools_with_two_vfs(self):
        # create two pools (PFs) with two VFs each
        self.pci_stats = stats.PciDeviceStats(
            objects.NUMATopology(), dev_filter=self.dev_filter
        )
        for pf_index in [1, 2]:
            for vf_index in [1, 2]:
                dev = objects.PciDevice(
                    compute_node_id=1,
                    vendor_id="dead",
                    product_id="beef",
                    address=f"0000:81:0{pf_index}.{vf_index}",
                    parent_addr=f"0000:81:0{pf_index}.0",
                    numa_node=0,
                    dev_type="type-VF",
                )
                self.pci_stats.add_device(dev)
                dev.extra_info = {'rp_uuid': getattr(uuids, f"pf{pf_index}")}

        # populate the RP -> pool mapping from the devices to its pools
        self.pci_stats.populate_pools_metadata_from_assigned_devices()

        # we have 2 pool and 4 devs in total
        self.num_pools = 2
        self.assertEqual(self.num_pools, len(self.pci_stats.pools))
        self.num_devs = 4
        self.assertEqual(
            self.num_devs, sum(pool["count"] for pool in self.pci_stats.pools)
        )

    def test_apply_asymmetric_allocation(self):
        self._create_two_pools_with_two_vfs()
        # ask for 3 VFs
        vf_req = objects.InstancePCIRequest(
            count=3,
            alias_name='a-vf',
            request_id=uuids.vf_req,
            spec=[
                {
                    "vendor_id": "dead",
                    "product_id": "beef",
                    "dev_type": "type-VF",
                }
            ],
        )

        # Simulate that placement returned an allocation candidate where 1 VF
        # is consumed from PF1 and two from PF2
        mapping = {
            # the VF is represented by the parent PF RP
            f"{uuids.vf_req}-0": [uuids.pf1],
            f"{uuids.vf_req}-1": [uuids.pf2],
            f"{uuids.vf_req}-2": [uuids.pf2],
        }
        # This should fit
        self.assertTrue(
            self.pci_stats.support_requests([vf_req], mapping)
        )
        # and when consumed the consumption from the pools should be in sync
        # with the placement allocation. So the PF2 pool is expected to
        # disappear as it is fully consumed and the PF1 pool should have
        # one free device.
        self.pci_stats.apply_requests([vf_req], mapping)
        self.assertEqual(1, len(self.pci_stats.pools))
        self.assertEqual(uuids.pf1, self.pci_stats.pools[0]['rp_uuid'])
        self.assertEqual(1, self.pci_stats.pools[0]['count'])

    def test_consume_asymmetric_allocation(self):
        self._create_two_pools_with_two_vfs()
        # ask for 3 VFs
        vf_req = objects.InstancePCIRequest(
            count=3,
            alias_name='a-vf',
            request_id=uuids.vf_req,
            spec=[
                {
                    "vendor_id": "dead",
                    "product_id": "beef",
                    "dev_type": "type-VF",
                    # Simulate that the scheduler already allocate a candidate
                    # and the mapping is stored in the request.
                    # In placement 1 VF is allocated from PF1 and two from PF2
                    "rp_uuids": ",".join([uuids.pf1, uuids.pf2, uuids.pf2])
                }
            ],
        )

        # So when the PCI claim consumes devices based on this request we
        # expect that nova follows what is allocated in placement.
        devs = self.pci_stats.consume_requests([vf_req])
        self.assertEqual(
            {"0000:81:01.0": 1, "0000:81:02.0": 2},
            collections.Counter(dev.parent_addr for dev in devs),
        )

    def test_consume_restricted_by_allocation(self):
        pf_req = objects.InstancePCIRequest(
            count=1,
            alias_name='a-dev',
            request_id=uuids.req1,
            spec=[
                {
                    "vendor_id": "dead",
                    "product_id": "beef",
                    "dev_type": "type-PF",
                    # Simulate that the scheduler already allocate a candidate
                    # and the mapping is stored in the request.
                    # The allocation restricts that we can only consume from
                    # PF3
                    "rp_uuids": ",".join([uuids.pf3])
                }
            ],
        )

        # Call consume. It always expects the allocated mapping to be stores
        # the in PCI request as it is always called from the compute side.
        consumed_devs = self.pci_stats.consume_requests([pf_req])
        # assert that the right device is consumed
        self.assertEqual([self.pf3], consumed_devs)
        # pf3 is not available in the pools anymore
        self.assertEqual(
            {uuids.pf1, uuids.pf2, uuids.pci1},
            {
                pool["rp_uuid"]
                for pool in self.pci_stats.pools
                if pool["count"] > 0
            },
        )


class PciDeviceVFPFStatsTestCase(test.NoDBTestCase):

    def setUp(self):
        super(PciDeviceVFPFStatsTestCase, self).setUp()
        device_spec = [
            jsonutils.dumps({"vendor_id": "8086", "product_id": "1528"}),
            jsonutils.dumps({"vendor_id": "8086", "product_id": "1515"}),
            jsonutils.dumps(
                {
                    "vendor_id": "15b3",
                    "product_id": "a2d6",
                    "remote_managed": "false",
                }
            ),
            jsonutils.dumps(
                {
                    "vendor_id": "15b3",
                    "product_id": "101e",
                    "remote_managed": "true",
                }
            ),
        ]
        self.flags(device_spec=device_spec, group='pci')
        self.pci_stats = stats.PciDeviceStats(objects.NUMATopology())

    def _create_pci_devices(self, vf_product_id=1515, pf_product_id=1528):
        self.sriov_pf_devices = []
        for dev in range(2):
            pci_dev = {
                'compute_node_id': 1,
                'address': '0000:81:00.%d' % dev,
                'vendor_id': '8086',
                'product_id': '%d' % pf_product_id,
                'status': 'available',
                'request_id': None,
                'dev_type': fields.PciDeviceType.SRIOV_PF,
                'parent_addr': None,
                'numa_node': 0
            }
            dev_obj = objects.PciDevice.create(None, pci_dev)
            dev_obj.child_devices = []
            self.sriov_pf_devices.append(dev_obj)

        # PF devices for remote_managed VFs.
        self.sriov_pf_devices_remote = []
        for dev in range(2):
            pci_dev = {
                'compute_node_id': 1,
                'address': '0001:81:00.%d' % dev,
                'vendor_id': '15b3',
                'product_id': 'a2d6',
                'status': 'available',
                'request_id': None,
                'dev_type': fields.PciDeviceType.SRIOV_PF,
                'parent_addr': None,
                'numa_node': 0,
                "capabilities": {"vpd": {
                    "card_serial_number": "MT2113X00000"}},
            }
            dev_obj = objects.PciDevice.create(None, pci_dev)
            dev_obj.child_devices = []
            self.sriov_pf_devices_remote.append(dev_obj)

        self.sriov_vf_devices = []
        for dev in range(8):
            pci_dev = {
                'compute_node_id': 1,
                'address': '0000:81:10.%d' % dev,
                'vendor_id': '8086',
                'product_id': '%d' % vf_product_id,
                'status': 'available',
                'request_id': None,
                'dev_type': fields.PciDeviceType.SRIOV_VF,
                'parent_addr': '0000:81:00.%d' % int(dev / 4),
                'numa_node': 0
            }
            dev_obj = objects.PciDevice.create(None, pci_dev)
            dev_obj.parent_device = self.sriov_pf_devices[int(dev / 4)]
            dev_obj.parent_device.child_devices.append(dev_obj)
            self.sriov_vf_devices.append(dev_obj)

        self.sriov_vf_devices_remote = []
        for dev in range(8):
            pci_dev = {
                'compute_node_id': 1,
                'address': '0001:81:10.%d' % dev,
                'vendor_id': '15b3',
                'product_id': '101e',
                'status': 'available',
                'request_id': None,
                'dev_type': fields.PciDeviceType.SRIOV_VF,
                'parent_addr': '0001:81:00.%d' % int(dev / 4),
                'numa_node': 0,
                "capabilities": {"vpd": {"card_serial_number": "MT2113X00000"}}
            }
            dev_obj = objects.PciDevice.create(None, pci_dev)
            dev_obj.parent_device = self.sriov_pf_devices_remote[int(dev / 4)]
            dev_obj.parent_device.child_devices.append(dev_obj)
            self.sriov_vf_devices_remote.append(dev_obj)

        self.vdpa_devices = []
        for dev in range(8):
            pci_dev = {
                'compute_node_id': 1,
                'address': '0000:82:10.%d' % dev,
                'vendor_id': '8086',
                'product_id': '%d' % vf_product_id,
                'status': 'available',
                'request_id': None,
                'dev_type': fields.PciDeviceType.VDPA,
                'parent_addr': '0000:81:00.%d' % int(dev / 4),
                'numa_node': 0
            }
            dev_obj = objects.PciDevice.create(None, pci_dev)
            dev_obj.parent_device = self.sriov_pf_devices[int(dev / 4)]
            dev_obj.parent_device.child_devices.append(dev_obj)
            self.vdpa_devices.append(dev_obj)

        list(map(self.pci_stats.add_device, self.sriov_pf_devices))
        list(map(self.pci_stats.add_device, self.sriov_vf_devices))
        list(map(self.pci_stats.add_device, self.vdpa_devices))
        list(map(self.pci_stats.add_device, self.sriov_pf_devices_remote))
        list(map(self.pci_stats.add_device, self.sriov_vf_devices_remote))

    def test_consume_VDPA_requests(self):
        self._create_pci_devices()
        pci_requests = [
            objects.InstancePCIRequest(
                count=8, spec=[{'dev_type': 'vdpa'}])]
        devs = self.pci_stats.consume_requests(pci_requests)
        self.assertEqual(8, len(devs))
        self.assertEqual('vdpa', devs[0].dev_type)
        free_devs = self.pci_stats.get_free_devs()
        # Validate that the parents of these devs has been removed
        # from pools.
        for dev in devs:
            self.assertNotIn(dev.parent_addr,
                             [free_dev.address for free_dev in free_devs])

    def test_consume_VF_requests(self):
        self._create_pci_devices()
        pci_requests = [objects.InstancePCIRequest(count=2,
                            spec=[{'product_id': '1515'}])]
        devs = self.pci_stats.consume_requests(pci_requests)
        self.assertEqual(2, len(devs))
        self.assertEqual(set(['1515']),
                            set([dev.product_id for dev in devs]))
        free_devs = self.pci_stats.get_free_devs()
        # Validate that the parents of these VFs has been removed
        # from pools.
        for dev in devs:
            self.assertNotIn(dev.parent_addr,
                             [free_dev.address for free_dev in free_devs])

    def test_consume_PF_requests(self):
        self._create_pci_devices()
        pci_requests = [objects.InstancePCIRequest(count=2,
                            spec=[{'product_id': '1528',
                                    'dev_type': 'type-PF'}])]
        devs = self.pci_stats.consume_requests(pci_requests)
        self.assertEqual(2, len(devs))
        self.assertEqual(set(['1528']),
                            set([dev.product_id for dev in devs]))
        free_devs = self.pci_stats.get_free_devs()
        # Validate that there are no free devices left, as when allocating
        # both available PFs, its VFs should not be available.
        self.assertEqual(0, len([d for d in free_devs
                                 if d.product_id == '1515']))

    def test_consume_VF_and_PF_requests(self):
        self._create_pci_devices()
        pci_requests = [objects.InstancePCIRequest(count=2,
                            spec=[{'product_id': '1515'}]),
                        objects.InstancePCIRequest(count=1,
                            spec=[{'product_id': '1528',
                                    'dev_type': 'type-PF'}])]
        devs = self.pci_stats.consume_requests(pci_requests)
        self.assertEqual(3, len(devs))
        self.assertEqual(set(['1528', '1515']),
                            set([dev.product_id for dev in devs]))

    def test_consume_VF_and_PF_requests_failed(self):
        self._create_pci_devices()
        pci_requests = [objects.InstancePCIRequest(count=5,
                            spec=[{'product_id': '1515'}]),
                        objects.InstancePCIRequest(count=1,
                            spec=[{'product_id': '1528',
                                    'dev_type': 'type-PF'}])]
        self.assertRaises(
            exception.PciDeviceRequestFailed,
            self.pci_stats.consume_requests,
            pci_requests,
        )

    def test_consume_VF_and_PF_same_product_id_failed(self):
        self._create_pci_devices(pf_product_id=1515)
        pci_requests = [objects.InstancePCIRequest(count=9,
                            spec=[{'product_id': '1515'}])]
        self.assertRaises(
            exception.PciDeviceRequestFailed,
            self.pci_stats.consume_requests,
            pci_requests,
        )

    def test_consume_PF_not_remote_managed(self):
        self._create_pci_devices()
        pci_requests = [objects.InstancePCIRequest(count=2,
                            spec=[{'product_id': '1528',
                                   'dev_type': 'type-PF',
                                   PCI_REMOTE_MANAGED_TAG: 'false'}])]
        devs = self.pci_stats.consume_requests(pci_requests)
        self.assertEqual(2, len(devs))
        self.assertEqual(set(['1528']),
                            set([dev.product_id for dev in devs]))
        free_devs = self.pci_stats.get_free_devs()
        # Validate that there are no free devices left with the
        # product ID under test, as when allocating both available
        # PFs, its VFs should not be available.
        self.assertEqual(0, len([d for d in free_devs
                                 if d.product_id == '1528']))

    def test_consume_VF_requests_remote_managed(self):
        self._create_pci_devices()
        pci_requests = [objects.InstancePCIRequest(count=2,
                            spec=[{PCI_REMOTE_MANAGED_TAG: 'true'}])]
        devs = self.pci_stats.consume_requests(pci_requests)
        self.assertEqual(2, len(devs))
        self.assertEqual(set(['101e']),
                            set([dev.product_id for dev in devs]))
        free_devs = self.pci_stats.get_free_devs()
        # Validate that the parents of these VFs has been removed
        # from pools.
        for dev in devs:
            self.assertNotIn(dev.parent_addr,
                             [free_dev.address for free_dev in free_devs])

    def test_consume_VF_requests_remote_managed_filtered(self):
        self._create_pci_devices()
        pci_requests = [objects.InstancePCIRequest(count=1,
                            spec=[{'product_id': '101e',
                                   PCI_REMOTE_MANAGED_TAG: 'false'}]),
                        objects.InstancePCIRequest(count=1,
                            spec=[{'product_id': '101e'}])]
        free_devs_before = self.pci_stats.get_free_devs()
        self.assertRaises(
            exception.PciDeviceRequestFailed,
            self.pci_stats.consume_requests,
            pci_requests,
        )
        free_devs_after = self.pci_stats.get_free_devs()
        self.assertEqual(free_devs_before, free_devs_after)

    def test_consume_VF_requests_remote_managed_mix(self):
        self._create_pci_devices()
        pci_requests = [objects.InstancePCIRequest(count=1,
                            spec=[{'product_id': '101e',
                                   PCI_REMOTE_MANAGED_TAG: 'true'}]),
                        objects.InstancePCIRequest(count=1,
                            spec=[{'product_id': '1515',
                                   PCI_REMOTE_MANAGED_TAG: 'false'}])]
        devs = self.pci_stats.consume_requests(pci_requests)
        self.assertEqual(2, len(devs))
        self.assertEqual(set(['101e', '1515']),
                            set([dev.product_id for dev in devs]))
        free_devs = self.pci_stats.get_free_devs()
        # Validate that the parents of these VFs has been removed
        # from pools.
        for dev in devs:
            self.assertNotIn(dev.parent_addr,
                             [free_dev.address for free_dev in free_devs])



# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime

import fixtures
from oslo_log import log as logging
from oslo_utils.fixture import uuidsentinel as uuids

from nova import conf
from nova import context as nova_context
from nova.tests import fixtures as nova_fixtures
from nova.tests.functional.libvirt import base

LOG = logging.getLogger(__name__)
CONF = conf.CONF


class RescueServerTestWithDeletedBaseImage(
    base.ServersTestBase
):

    api_major_version = 'v2.1'
    microversion = '2.87'

    # BUG #2002606
    # Rescue in stable rescue mode with deleted original image must be
    # successful
    def setUp(self):
        super(RescueServerTestWithDeletedBaseImage, self).setUp()

        self.ctxt = nova_context.get_admin_context()

        # to create bfv server with libvirt driver
        self.cinder = self.useFixture(nova_fixtures.CinderFixture(self))
        self.useFixture(nova_fixtures.OSBrickFixture())
        # disk.rescue image_create ignoring privsep
        self.useFixture(fixtures.MockPatch(
            'nova.virt.libvirt.imagebackend._update_utime_ignore_eacces'))

        # dir to create 'unrescue.xml'
        def fake_path(_self, *args, **kwargs):
            return CONF.instances_path

        self.useFixture(fixtures.MonkeyPatch(
            'nova.virt.libvirt.utils.get_instance_path', fake_path))

    def _create_test_images(self):
        timestamp = datetime.datetime(2021, 1, 2, 3, 4, 5)
        base_image = {
            'id': uuids.base_image,
            'name': 'base_image',
            'created_at': timestamp,
            'updated_at': timestamp,
            'deleted_at': None,
            'deleted': False,
            'status': 'active',
            'is_public': False,
            'container_format': 'ova',
            'disk_format': 'vhd',
            'size': '74185822',
            'min_ram': 0,
            'min_disk': 0,
            'protected': False,
            'visibility': 'public',
            'tags': [],
            'properties': {}
        }
        rescue_stable_image = {
            'id': uuids.rescue_stable_image,
            'name': 'base_image',
            'created_at': timestamp,
            'updated_at': timestamp,
            'deleted_at': None,
            'deleted': False,
            'status': 'active',
            'is_public': False,
            'container_format': 'ova',
            'disk_format': 'vhd',
            'size': '74185822',
            'min_ram': 0,
            'min_disk': 0,
            'protected': False,
            'visibility': 'public',
            'tags': [],
            'properties': {
               'hw_rescue_bus': 'scsi',
               'hw_rescue_device': 'cdrom'
            }
        }
        self.glance.create(self.ctxt, base_image)
        self.glance.create(self.ctxt, rescue_stable_image)
        return base_image, rescue_stable_image

    def test_stable_rescue_server_local_disk(self):
        self.start_compute()

        # create a local server with base image
        base_image, rescue_stable_image = self._create_test_images()
        server = self._create_server(image_uuid=uuids.base_image,
            networks='none')
        # instance.image_ref exists
        self.assertEqual(base_image['id'], server['image']['id'])

        # Delete base image
        self.glance.delete(self.ctxt, base_image['id'])

        rescue_req = {
            'rescue': {
                'rescue_image_ref': rescue_stable_image['id']
            }
        }
        # Rescue in stable device mode successful
        res = self.api.api_post('/servers/%s/action' % server['id'],
            rescue_req)
        self.assertEqual(200, res.status)

        # BUG #2002606
        # This test fails and server never reach RESCUE state.
        # It is expected behavior when server in stable rescue mode cannot
        # find original glance image.
        #
        # FIX with fallback:
        # except exception.ImageNotFound:
        #    image_meta = instance.image_meta
        # leads to passing this test

        self._wait_for_state_change(server, 'RESCUE')

    def test_stable_rescue_server_bfv(self):
        self.start_compute()

        # create a boot from volume server with libvirt driver
        base_image_not_used, rescue_stable_image = self._create_test_images()
        server_request = self._build_server(networks=[])
        server_request.pop('imageRef')
        server_request['block_device_mapping_v2'] = [{
            'boot_index': 0,
            'uuid': nova_fixtures.CinderFixture.IMAGE_BACKED_VOL,
            'source_type': 'volume',
            'destination_type': 'volume'}]
        server = self.api.post_server({'server': server_request})
        server = self._wait_for_state_change(server, 'ACTIVE')
        # instance.image_ref is missing, attached volume exists
        self.assertEqual('', server['image'])
        self.assertEqual(1,
           len(server['os-extended-volumes:volumes_attached']))

        # Delete all glance images except rescue image
        ids = []
        for o in self.glance.images.keys():
            ids.append(o)
        for o in ids:
            if o != rescue_stable_image['id']:
                self.glance.delete(self.ctxt, o)
        self.assertEqual(1, len(self.glance.images.keys()))

        rescue_req = {
            'rescue': {
                'rescue_image_ref': rescue_stable_image['id']
            }
        }
        # Rescue in stable device mode successful
        res = self.api.api_post('/servers/%s/action' % server['id'],
            rescue_req)
        self.assertEqual(200, res.status)

        # BUG #2002606 does not affect server with bfv root
        # disks unless server.image_ref defined for bfv server.

        self._wait_for_state_change(server, 'RESCUE')



# Copyright 2014 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from nova.api.validation import parameter_types


create_backup = {
    'type': 'object',
    'properties': {
        'createBackup': {
            'type': 'object',
            'properties': {
                'name': parameter_types.name,
                'backup_type': {
                    'type': 'string',
                },
                'rotation': parameter_types.non_negative_integer,
                'metadata': parameter_types.metadata,
            },
            'required': ['name', 'backup_type', 'rotation'],
            'additionalProperties': False,
        },
    },
    'required': ['createBackup'],
    'additionalProperties': False,
}


create_backup_v20 = copy.deepcopy(create_backup)
create_backup_v20['properties'][
    'createBackup']['properties']['name'] = (
        parameter_types.name_with_leading_trailing_spaces)


create_backup_response = {
    'type': 'null',
}

create_backup_response_v245 = {
    'type': 'object',
    'properties': {
        'image_id': {'type': 'string', 'format': 'uuid'},
    },
    'required': ['image_id'],
    'additionalProperties': False,
}



# Copyright 2011-2012 OpenStack Foundation
# All Rights Reserved.
# Copyright 2013 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from webob import exc

from nova.api.openstack.compute.schemas import cells as schema
from nova.api.openstack import wsgi
from nova.api import validation

_removal_reason = """\
This API only works in a Cells v1 deployment, which was deprecated in the
16.0.0 (Pike) release. It is not used with Cells v2, which is required
beginning in the 15.0.0 (Ocata) release.
It was removed in the 20.0.0 (Train) release.
"""


class CellsController(wsgi.Controller):
    """(Removed) Controller for Cell resources.

    This was removed during the Train release in favour of cells v2.
    """

    @wsgi.expected_errors(410)
    @wsgi.removed('20.0.0', _removal_reason)
    @validation.query_schema(schema.index_query)
    def index(self, req):
        raise exc.HTTPGone()

    @wsgi.expected_errors(410)
    @wsgi.removed('20.0.0', _removal_reason)
    @validation.query_schema(schema.detail_query)
    def detail(self, req):
        raise exc.HTTPGone()

    @wsgi.expected_errors(410)
    @wsgi.removed('20.0.0', _removal_reason)
    @validation.query_schema(schema.info_query)
    def info(self, req):
        raise exc.HTTPGone()

    @wsgi.expected_errors(410)
    @wsgi.removed('20.0.0', _removal_reason)
    @validation.query_schema(schema.capacities_query)
    def capacities(self, req, id=None):
        raise exc.HTTPGone()

    @wsgi.expected_errors(410)
    @wsgi.removed('20.0.0', _removal_reason)
    @validation.query_schema(schema.show_query)
    def show(self, req, id):
        raise exc.HTTPGone()

    @wsgi.expected_errors(410)
    @wsgi.removed('20.0.0', _removal_reason)
    def delete(self, req, id):
        raise exc.HTTPGone()

    @wsgi.expected_errors(410)
    @wsgi.removed('20.0.0', _removal_reason)
    @validation.schema(schema.create)
    def create(self, req, body):
        raise exc.HTTPGone()

    @wsgi.expected_errors(410)
    @wsgi.removed('20.0.0', _removal_reason)
    @validation.schema(schema.update)
    def update(self, req, id, body):
        raise exc.HTTPGone()

    @wsgi.expected_errors(410)
    @wsgi.removed('20.0.0', _removal_reason)
    @validation.schema(schema.sync_instances)
    def sync_instances(self, req, body):
        raise exc.HTTPGone()
